
__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2015, RADICAL@Rutgers"
__license__   = "MIT"


# ------------------------------------------------------------------------------

from watcher_mem  import WatcherMem
from watcher_sto  import WatcherSto
from watcher_cpu  import WatcherCPU
from watcher_sys  import WatcherSys


# ------------------------------------------------------------------------------




#!/usr/bin/env python


import os
import sys
import time   
import pymongo

import radical.utils as ru

ATTEMPTS = 10

# ------------------------------------------------------------------------------
#
def work (master_id, worker_id) :
    """
    Connect to mongodb at the given location, grab a work item from it, invoke
    the mb calculation for the given region, and push the results back.

    The path element of the master_id (which is considered to be a URL) is used
    as collection ID, the client ID will specify the worker document.
    """


    [host, port, dbname, cname, _, _, _] = ru.split_dburl (master_id)
    [host, port, dbname, cname, _, _, _] = ru.split_dburl (master_id)

    print "  %s:%d" % (host, port)
    db_client  = pymongo.MongoClient (host=host, port=port)
    database   = db_client[dbname]
    collection = database[cname]
    docs = collection.find ()

  # for i in db_client.database_names  () : print "  %s" % i
  # for i in database.collection_names () : print "     %s" % i
  # print "      %d docs" % docs.count()

    print 'master id : %s' % master_id
    print 'worker id : %s' % worker_id
    print 'host      : %s' % host
    print 'port      : %s' % port
    print 'dbname    : %s' % dbname
    print 'cname     : %s' % cname
    print 'worker_id : %s' % worker_id 
    print 'works     : %s' % collection.find ().count()

    attempts = 0

    while attempts < ATTEMPTS :
        work = collection.find_one ({'worker_id': int(worker_id),
                                     'type'     : 'work'})
        if  work :
            break
        else :
            print 'waiting for work'
            time.sleep (1)

    if  not work :
        print "cannot find work for %s : %s" % (master_id, worker_id)
        sys.exit (0) # silent exit


    print 'work      : %s' % str(work)

    result = mandelbrot_calc (work)

    print 'publish result'
    collection.insert (result)

    db_client.disconnect ()


# ------------------------------------------------------------------------------
#
def mandelbrot_calc (work) :
    """    
    compute the mandelbrot (sub)set with the given parameters.
    `work` is considered to be a dict with the following keys::

        xmin, ymin :  lower left corner in complex plane
        xmax, ymax :  upper right corner in complex plane
        xpix, ypix :  sub-picture resolution 
        iters      :  max iterations

    """

    start = time.time()

    if  'minx'  not in work : usage ("No 'minx'   in work '%s'" % work)
    if  'miny'  not in work : usage ("No 'miny'   in work '%s'" % work)
    if  'maxx'  not in work : usage ("No 'maxx'   in work '%s'" % work)
    if  'maxy'  not in work : usage ("No 'maxy'   in work '%s'" % work)
    if  'pixx'  not in work : usage ("No 'pixx'   in work '%s'" % work)
    if  'pixy'  not in work : usage ("No 'pixy'   in work '%s'" % work)
    if  'subx'  not in work : usage ("No 'subx'   in work '%s'" % work)
    if  'suby'  not in work : usage ("No 'suby'   in work '%s'" % work)
    if  'iters' not in work : usage ("No 'iters'  in work '%s'" % work)

    minx  = work['minx']
    miny  = work['miny']
    maxx  = work['maxx']
    maxy  = work['maxy']
    pixx  = work['pixx']
    pixy  = work['pixy']
    subx  = work['subx']
    suby  = work['suby']
    iters = work['iters']

  # print "minx  : %s " % minx
  # print "miny  : %s " % miny
  # print "maxx  : %s " % maxx
  # print "maxy  : %s " % maxy
  # print "pixx  : %s " % pixx
  # print "pixy  : %s " % pixy
  # print "subx  : %s " % subx
  # print "suby  : %s " % suby
  # print "iters : %s " % iters

    data = list()

    for y in range (0, pixy) :
    
        zy = y * (maxy - miny) / (pixy - 1) + miny

        line = list()
    
        for x in range (0, pixx) :
    
            zx = x * (maxx - minx) / (pixx - 1)  + minx
            z  = zx + zy * 1j
            c  = z
    
            for i in range (0, iters) :
                if  abs(z) > 2.0 :
                    break 
    
                z = z * z + c

            line.append (i)

        data.append (line)


    result = {'worker_id' : int(worker_id), 
              'type'      : 'result' ,
              'pixx'      : pixx, 
              'pixy'      : pixy, 
              'subx'      : subx, 
              'suby'      : suby, 
              'data'      : data, 
              'time'      : time.time()-start,
              'host'      : os.popen('hostname -f').read().strip()}

    return result


# ------------------------------------------------------------------------------
#
def usage (msg=None) :
    """
    print usage.  If message is given, we exit with that error message
    otherwise we exit peacefully .
    """

    if  msg :
        sys.stderr.write ('\n    Error: %s' % msg)

    print """

    usage :

        python mandelbrot_worker.py
               --master_id=mongodb://host.net:port/path
               --worker_id=3

    """

    if  msg :
        sys.exit (1)

    else :
        sys.exit (0)


# ------------------------------------------------------------------------------
#
if __name__ == '__main__' :
    """
    parse arguments, and trigger workload.
    This slave is basically stateless: it either works or does not.  It will
    grab a *single* work item from  mongodb and calculate the mb subset
    specified within, and then will store the result back at the same location,
    and exits.

    """

    master_id = None
    worker_id = None

    for arg in sys.argv[1:] :

        if  arg == '--help' :
            usage ()

        else :
            if  not '=' in arg :
                usage ("invalid argument '%s'" % arg)

            key, val = arg.split ('=', 1)

            if  key == '--master_id' :
                master_id = val

            elif key == '--worker_id' :
                worker_id = val

            else :
                # ignore other parameters, those are used for profile indexing
                pass
              # usage ("parameter '%s' not supported" % arg)


    if  not master_id or not worker_id :
        usage ("need master_id and worker_id to operate (%s)" % sys.argv)

    
    work (master_id, worker_id)

    sys.exit (0)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2015, RADICAL@Rutgers"
__license__   = "MIT"


import os
import time
import threading

from ..utils import timestamp

# number of samples per second
_SAMPLE_RATE = 1

# ------------------------------------------------------------------------------
#
class WatcherBase (threading.Thread) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self, pid):

        threading.Thread.__init__(self)
        self._terminate = threading.Event()
        self._pid       = pid
        self._data      = dict()
        self._config    = dict()

        self.start ()


    # --------------------------------------------------------------------------
    #
    # This is the API which watcher implementations overload.
    #
    def _pre_process  (self)       : pass
    def _post_process (self)       : pass
    def _sample       (self, now)  : pass
    def _finalize     (self, info) : pass


    # --------------------------------------------------------------------------
    #
    def finalize (self, info) :

        try:
            self._finalize (info)

        except Exception as e:
            
            import traceback
            traceback.print_exc()

            print "Exception in finalize: %s" % e


    # --------------------------------------------------------------------------
    #
    def stop (self) :

        self._terminate.set()


    # --------------------------------------------------------------------------
    #
    def run (self) :

        try:

            self._config['sample_rate'] = float(os.environ.get('RADICAL_SYNAPSE_SAMPLERATE', _SAMPLE_RATE))

            # sample rate is samples per minute
            sleeptime = 1/self._config['sample_rate']

            self._pre_process(self._config)

            while not self._terminate.is_set():

                now = timestamp()
                self._sample(now)
                
                time.sleep (sleeptime)

            self._post_process()

          # print " ----- %s -----" % self.__class__
          # pprint.pprint (self._data)

        except Exception as e:

            import traceback
            traceback.print_exc()

            print "Exception in watcher: %s" % e


    # --------------------------------------------------------------------------
    #
    def get_data (self) :

        return self._data



# ------------------------------------------------------------------------------





import radical.synapse

# ------------------------------------------------------------------------------
#

NOTHING = None
PROFILE = 'synapse_profile'
EMULATE = 'synapse_emulate'

# ------------------------------------------------------------------------------
#
def synapsify (command, mode=NOTHING) :
    """
    Uses synapse to transform the given command in three possible ways,
    depending on mode::

        NOTHING: 
            empty transormation -- duh!  This only exists to avoid case
            switches in the application code

        PROFILE:
            the command is run under a profile, which will, duh, profile the
            application run, and store the results in the synapse MongoDB
            database.

        EMULATE:
            the command is replaces with a synapse emulation -- the emulation
            parameters are pulled from the synapse MongoDB database.  If those
            parameters cannot be found, an exception is raised.
    """

    if  mode == NOTHING :
        radical.synapse._logger.debug ("synapsify in : %s" % command)
        radical.synapse._logger.debug ("synapsify out: %s" % command)
        return command


    if  mode  == PROFILE :
        ret = "radical-synapse-profile.py '%s'" % command
        radical.synapse._logger.debug ("synapsify in : %s" % command)
        radical.synapse._logger.debug ("synapsify out: %s" % ret)
        return ret


    if  mode  == EMULATE :
        ret = "radical-synapse-emulate.py '%s'" % command
        radical.synapse._logger.debug ("synapsify in : %s" % command)
        radical.synapse._logger.debug ("synapsify out: %s" % ret)
        return ret


# ------------------------------------------------------------------------------




#!/usr/bin/env python


import os
import sys
import time
import numpy
import pprint        as pp
import radical.utils as ru

import synapse as rs

# import pudb 
# pudb.set_interrupt_handler ()

iters = 10
host  = os.getenv ('HOST', os.popen ('hostname | cut -f 1 -d . | xargs echo -n').read ())

# ------------------------------------------------------------------------------
#
def mandel (x, y, z) :

  # print 'mandel: %s %s %s' % (x, y, z)

    start = time.time ()

    x_pixels  = x
    y_pixels  = y
    iter_max  = z
    x_min     = -2.0
    x_max     = +1.0
    y_min     = -1.5
    y_max     = +1.5
    threshold = 1
    rgb_image = []

    for x in range (x_pixels) :

      # print x

        cx = x * (x_max - x_min) / x_pixels + x_min

        rgb_row = []

        for y in range (y_pixels):

            cy = y * (y_max - y_min) / (y_pixels - 1) + x_min

            c = complex (cx, cy)
            z = 0

            for i in range (iter_max) :
                
                if abs (z) > threshold: 
                    break 

                z = z * z + c 

            rgb_row.append (["%3d" % i, "%3d" % i, "%3d" % i])

        rgb_image.append (rgb_row)

    with open ("/tmp/mb.dat", "w") as f :
        f.write (str(rgb_image))

    stop = time.time ()

    return stop - start


# ------------------------------------------------------------------------------
#

def main (cfg_list) :

    n2h = rs.number_to_human

    for cfg in cfg_list :
        x       = int  (cfg[ 2])
        y       = int  (cfg[ 3])
        z       = int  (cfg[ 4])
        stress  = 0
        compute = 0
        util    = 0
        eff     = 0
        load    = 0
        memory  = 0
        storage = 0

        iter_1_time = list()
        iter_1_cpu  = list()
        iter_1_mem  = list()
        iter_1_io   = list()
        iter_1_util = list()
        iter_1_eff  = list()
        iter_1_sys  = list()

        for iter in range (iters) :

            info_run, ret, out = rs.profile (mandel, x, y, z)

            if not 'utilization' in info_run['cpu'] : info_run['cpu']['utilization' ] = 0
            if not 'efficiency'  in info_run['cpu'] : info_run['cpu']['efficiency'  ] = 0
            if not 'load'        in info_run['cpu'] : info_run['cpu']['load'        ] = 0
            if not 'ops'         in info_run['cpu'] : info_run['cpu']['ops'         ] = 0
            if not 'max'         in info_run['mem'] : info_run['mem']['max'         ] = 0
            if not 'write'       in info_run['io']  : info_run['io']['write'        ] = 0
        
            load_compute = int(float(info_run['cpu']['ops' ]) / (1024*1024) / 8)
            load_memory  = int(float(info_run['mem']['max' ]) / (1024*1024))
            load_storage = int(float(info_run['io']['write']) / (1024*1024))
            
            load_compute = int(float(info_run['cpu']['ops' ]) / (1024*1024) / 8)
            load_memory  = int(float(info_run['mem']['max' ]) / (1024*1024))
            load_storage = int(float(info_run['io']['write']) / (1024*1024))
            
            load_id  = 'RUN.%04d' % x
            output   = '%-10s %15s %5d %5d %5d %6d %8.2f --------- --------- --------- %9.2f %0.2f %0.2f %5.2f %9.2f %9.2f' % \
                       (host, load_id, x, y, z, stress,
                        ret, 
                        load_compute, 
                        info_run['cpu']['utilization'], 
                        info_run['cpu']['efficiency'], 
                        info_run['sys']['load'],
                        load_memory, 
                        load_storage
                       )
            print output

            iter_1_time .append(float(ret))
            iter_1_cpu  .append(float(load_compute))
            iter_1_mem  .append(float(load_memory))
            iter_1_io   .append(float(load_storage))
            iter_1_util .append(float(info_run['cpu']['utilization']))
            iter_1_eff  .append(float(info_run['cpu']['efficiency']))
            iter_1_sys  .append(float(info_run['sys']['load']))

        iter_1_mean_time = numpy.mean (numpy.array (iter_1_time))
        iter_1_mean_cpu  = numpy.mean (numpy.array (iter_1_cpu ))
        iter_1_mean_mem  = numpy.mean (numpy.array (iter_1_mem ))
        iter_1_mean_io   = numpy.mean (numpy.array (iter_1_io  ))
        iter_1_mean_util = numpy.mean (numpy.array (iter_1_util))
        iter_1_mean_eff  = numpy.mean (numpy.array (iter_1_eff ))
        iter_1_mean_sys  = numpy.mean (numpy.array (iter_1_sys ))
    
        iter_1_std_time  = numpy.std  (numpy.array (iter_1_time))
        iter_1_std_cpu   = numpy.std  (numpy.array (iter_1_cpu ))
        iter_1_std_mem   = numpy.std  (numpy.array (iter_1_mem ))
        iter_1_std_io    = numpy.std  (numpy.array (iter_1_io  ))
        iter_1_std_util  = numpy.std  (numpy.array (iter_1_util))
        iter_1_std_eff   = numpy.std  (numpy.array (iter_1_eff ))
        iter_1_std_sys   = numpy.std  (numpy.array (iter_1_sys ))
        
        load_id  = 'MEAN_RUN.%04d' % x
        mean     = '%-10s %15s %5d %5d %5d %5d %9.2f --------- --------- --------- %9.2f %.2f %.2f %5.2f %9.2f %9.2f' % \
                   (host, load_id, x, y, z, stress,
                    iter_1_mean_time, 
                    iter_1_mean_cpu,
                    iter_1_mean_util, 
                    iter_1_mean_eff, 
                    iter_1_mean_sys,
                    iter_1_mean_mem, 
                    iter_1_mean_io
                   )
        print mean  
    
        load_id  = 'STD_RUN.%04d' % x
        std      = '%-10s %15s %5d %5d %5d %5d %9.2f --------- --------- --------- %9.2f %.2f %.2f %5.2f %9.2f %9.2f' % \
                   (host, load_id, x, y, z, stress, 
                    iter_1_std_time, 
                    iter_1_std_cpu,
                    iter_1_std_util, 
                    iter_1_std_eff, 
                    iter_1_std_sys,
                    iter_1_std_mem, 
                    iter_1_std_io
                   )
        print std
    
        load_id  = 'STAT_RUN.%04d' % x
        stat     = '%-10s %15s %5d %5d %5d %5d %9.2f %9.2f --------- --------- --------- --------- --------- --------- %9.2f %9.2f %.2f %.2f %.2f %.2f %5.2f %5.2f %9.2f %9.2f %9.2f %9.2f' % \
                   (host, load_id, x, y, z, stress, 
                    iter_1_mean_time, 
                    iter_1_std_time, 
                    iter_1_mean_cpu,
                    iter_1_std_cpu,
                    iter_1_mean_util, 
                    iter_1_std_util, 
                    iter_1_mean_eff, 
                    iter_1_std_eff, 
                    iter_1_mean_sys,
                    iter_1_std_sys,
                    iter_1_mean_mem, 
                    iter_1_std_mem, 
                    iter_1_mean_io,
                    iter_1_std_io
                   )
        print stat  
    
# ------------------------------------------------------------------------------
#
#
cfg_list = list ()

if  not len(sys.argv) > 1 :
    print "\n\tusage: %s <cfg_file> \n\n" % sys.argv[0]
    sys.exit (-1)

with open (sys.argv[1]) as fin :
    lines = fin.readlines ()

for line in lines :
    line = line.replace ('\n', '')
    re = ru.ReString (line)
    if  re // r'(MEAN_PRO)' :
        print line
        cfg_list.append (line.split())

main (cfg_list)

  





from ._atoms    import atom_compute_asm 
from ._atoms    import atom_compute
from ._atoms    import atom_time
from ._atoms    import atom_memory
from ._atoms    import atom_storage 
from ._atoms    import atom_network 

from .constants import UNKNOWN, TIME, COMPUTE, MEMORY, STORAGE, NETWORK

from .base      import AtomBase
from .timer     import Time
from .compute   import Compute
from .memory    import Memory
from .storage   import Storage
from .network   import Network




#!/usr/bin/env python


import os
import sys
import math
import time
import pymongo
from   PIL import Image

import radical.utils as ru

# ------------------------------------------------------------------------------
#
def usage (msg=None) :
    """
    print usage.  If message is given, we exit with that error message
    otherwise we exit peacefully .
    """

    if  msg :
        sys.stderr.write ('\n    Error: %s' % msg)

    print """

    usage :

        python mandelbrot_master.py
               --master_id=mongodb://host.net:port/path
               --num_workers=4
               --mb_size=1024
               --mb_depth=128

        or

        python mandelbrot_master.py
               -i=mongodb://host.net:port/path
               -n=4
               -s=1024
               -d=1024

        ID is mandatory.
    """

    if  msg :
        sys.exit (1)

    else :
        sys.exit (0)


# ------------------------------------------------------------------------------
#
def main (master_id, num_workers, mb_size, mb_depth) :

    [host, port, dbname, cname, _, _, _] = ru.split_dburl (master_id)

    db_client  = pymongo.MongoClient (host=host, port=port)
    database   = db_client[dbname]
    collection = database[cname]

  # collection.remove ()
  
    subsx   = int(math.sqrt(num_workers))
    subsy   = int(math.sqrt(num_workers))
    
    minx    = -2.0
    maxx    =  1.0
    miny    = -1.5
    maxy    =  1.5
    pixx    =  mb_size  # divisible by subsx
    pixy    =  mb_size  # divisible by subsy
    iters   =  mb_depth
    image   =  Image.new ("RGB", (pixx, pixy))
    
    stepx   = (maxx-minx) / subsx
    stepy   = (maxy-miny) / subsy
    
    spixx   =  pixx / subsx
    spixy   =  pixy / subsy

    # initialize white pic
    for     x in range (0, pixx) :
        for y in range (0, pixy) :
            image.putpixel ((x, y), (255, 255, 255))
    image.save  ("%s/mandel.png" % os.environ['HOME'], "PNG")


    workers = dict()
    
    # subdivide and publish work items
    for   subx in range (0, subsx) :
      for suby in range (0, subsy) :
    
          worker_idx = subx * subsx + suby
          worker_doc = {'worker_id' : int(worker_idx), 
                        'type'      : 'work',
                        'minx'      : float(minx + subx * stepx),
                        'miny'      : float(miny + suby * stepy),
                        'maxx'      : float(minx + subx * stepx + stepx),
                        'maxy'      : float(miny + suby * stepy + stepy),
                        'pixx'      : spixx,
                        'pixy'      : spixy,
                        'subx'      : subx,
                        'suby'      : suby,
                        'iters'     : int(iters)
                        }
  
          workers[worker_idx] = worker_doc
  
          collection.insert (workers[worker_idx])


    # wait for and evaluate results
    done    = list()

    while len(done) != len(workers) :

        print "waiting (%s/%s)" % (len(done), len(workers))

        results = collection.find ({'type' : 'result'})
        active  = False

        for result in results :

            worker_id = result['worker_id']
            pixx      = result['pixx']
            pixy      = result['pixy']
            subx      = result['subx']
            suby      = result['suby']

            if  worker_id not in done :

                active = True
                data   = result['data']

                for     x in range (0, pixx) :
                    for y in range (0, pixy) :
                        i = data[x][y]
                        image.putpixel ((spixx*subx+y, spixy*suby+x), (i/4, i/4, i))

                done.append (worker_id)
                image.save  ("%s/mandel.png" % os.environ['HOME'], "PNG")

        if  not active:
            time.sleep (1)

    print "done    (%s/%s)" % (len(done), len(workers))

    db_client.disconnect ()



# ------------------------------------------------------------------------------
#
if __name__ == '__main__' :

    """
    parse arguments, and trigger action workload.
    """

    master_id   = None
    num_workers = 1
    mb_size     = 1024
    mb_depth    = 1024

    for arg in sys.argv[1:] :

        if  arg == '--help' :
            usage ()

        else :
            if  not '=' in arg :
                usage ("invalid argument '%s'" % arg)

            key, val = arg.split ('=', 1)

            if  key == '--master_id' or \
                key == '-i'          :
                master_id = val

            elif key == '--num_workers' or \
                 key == '-n'            :
                num_workers = int(val)

            elif key == '--mb_size' or \
                 key == '-s'        :
                mb_size = int(val)

            elif key == '--mb_depth' or \
                 key == '-d'         :
                mb_depth = int(val)

            else :
                usage ("parameter '%s' not supported" % arg)


    if  not master_id :
        usage ("need master_id to operate (%s)" % sys.argv)

    main (master_id, num_workers, mb_size, mb_depth)
    





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


import radical.utils as ru


from   _atoms    import atom_network
from   base      import AtomBase
from   constants import NETWORK

# ------------------------------------------------------------------------------
#
class Network (AtomBase):
    """
    This Network Synapse emulates a network workload.  It operates in either one
    of two modes: receive or send.  On receive, the synapse will listen on
    a given port (forever), and on any incoming connection will read data until
    the connection is closed from the sending end.  All data are immediately
    discarded.  Reading happens in 1MB chunks.
    On sending, the synapse will (forever) attempt to contact a receiver given 
    port, will send the given amount of data in 1MB chunk, and will then close
    the connection.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self): 

        AtomBase.__init__ (self, NETWORK)


    # --------------------------------------------------------------------------
    #
    def _verify(self, vals): 

        self._proc = None

        assert('type' in vals)
        assert('mode' in vals)
        assert('port' in vals)

        typ  =     vals['type']
        mode =     vals['mode']
        port = int(vals['port'])
        host =     vals.get('size')
        size = int(vals.get('size', 1))

        assert(typ == 'client' or host)


    # --------------------------------------------------------------------------
    #
    def _emulate (self, vals):

        typ  =     vals['type']
        mode =     vals['mode']
        port = int(vals['port'])
        host =     vals.get('size')
        size = int(vals.get('size', 1))

        try:
            atom_network (typ, mode, host, port, size)

        except Exception as e:
            print "net atom error: %s" % e
            ru.cancel_main_thread()


#-------------------------------------------------------------------------------





import os
import sys
import time
import pprint
import threading
import radical.utils   as ru
import subprocess      as sp
import multiprocessing as mp

import watcher as rsw
import utils   as rsu
import atoms   as rsa

# import pudb
# pudb.set_interrupt_handler ()


# ------------------------------------------------------------------------------
#
LOAD = int (os.environ.get ('LOAD', '0'))

if 'darwin' not in sys.platform.lower():
  # LOAD_CMD    = "top -b -n1 | head -1  |       cut -f 4 -d :         | cut -f 1 -d ,"
    LOAD_CMD    = "top -b -n1 | head -n1 | rev | cut -f 3 -d \  | rev  | sed -e 's/,//'"
else:
    LOAD_CMD    = "top -l 1 | head -n 3 | tail -n 1 | cut -f 3 -d ' ' | cut -f 1 -d ','"



# ------------------------------------------------------------------------------
#
def execute (command, *args, **kwargs):

    if callable (command):
        cmd_str = "%s %s %s" % (command.__name__, str (args), str(kwargs))

    else:
        cmd_str = command


    print "execute: %s" % cmd_str

    info = {'cmd' : cmd_str}


    # start stress, get it spinning for one min to get a confirmed load
    # measurement, then run our own load, then kill stress.
    if  LOAD > 0:
        rsu.logger.info ("creating system load %s" % LOAD)
        os.popen ("killall -9 stress 2>&1 > /dev/null")
        os.popen ('stress --cpu %s &' % LOAD)
        time.sleep (60)

    load_1 = float(os.popen (LOAD_CMD).read())
    start  = rsu.timestamp()

    os.environ['_RADICAL_SYNAPSE_PROFILED'] = 'TRUE'

    # run the function/command in a separate process
    if callable (command):

        proc = mp.Process (target = command,
                           args   = args,
                           kwargs = kwargs)
        proc.start ()

    else:

        proc = sp.Popen (command.split(),
                        stdout = sp.PIPE,
                        stderr = sp.STDOUT)

    if callable (command):

        proc.join()
        out = ""
        ret = None

    else:
        out = proc.communicate()[0]
        ret = proc.returncode

    stop   = rsu.timestamp()
    load_2 = float(os.popen (LOAD_CMD).read())

    info['time'] = dict()
    info['time']['start'] = rsu.time_zero()
    info['time']['real']  = stop-start

    info['cpu']  = dict()
    info['cpu']['load']   = max(load_1, load_2)
   
    if  LOAD > 0:
        rsu.logger.info ("stopping system load")
        os.popen ("killall -9 stress 2>&1 > /dev/null")
        rsu.logger.info ("stopped  system load")

    return info, ret, out


# ------------------------------------------------------------------------------
#
def profile (command, *args, **kwargs) :

    if callable (command):
        cmd_str = "%s %s %s" % (command.__name__, str (args), str(kwargs))

    else:
        cmd_str = command


    print "profile: %s" % cmd_str

    if '_RADICAL_SYNAPSE_EMULATED' in os.environ:
        cmd_str = os.environ.get ('_RADICAL_SYNAPSE_EMULATEE', cmd_str)
        print 'using emulated command name: %s' % cmd_str

    info = {'cmd' : cmd_str}

    # start stress, get it spinning for one min to get a confirmed load
    # measurement, then run our own load, then kill stress.
    if  LOAD > 0:
        rsu.logger.info ("creating system load %s" % LOAD)
        os.popen ("killall -9 stress 2>&1 > /dev/null")
        os.popen ('stress --cpu %s &' % LOAD)
        time.sleep (60)

    load_1 = float(os.popen (LOAD_CMD).read())
    start  = rsu.timestamp()

    os.environ['_RADICAL_SYNAPSE_PROFILED'] = 'TRUE'

    # run the profiled function/command in a separate process
    if callable (command):

        proc = mp.Process (target = command,
                           args   = args,
                           kwargs = kwargs)
        proc.start ()

    else:

        proc = sp.Popen (command.split(),
                        stdout = sp.PIPE,
                        stderr = sp.STDOUT)

    watch_mode = os.environ.get('RADICAL_SYNAPSE_WATCHMODE', 'full').lower()
    watchers   = list()

    if watch_mode == 'full':
        watchers.append (rsw.WatcherCPU (proc.pid))
        watchers.append (rsw.WatcherSto (proc.pid))
        watchers.append (rsw.WatcherMem (proc.pid))

    # watchmode 'basic'
    watchers.append (rsw.WatcherSys (proc.pid))


    if callable (command):

        proc.join()
        out = ""
        ret = None

    else:
        out = proc.communicate()[0]
        ret = proc.returncode

    stop = rsu.timestamp()

    info['time'] = dict()
    info['time']['start'] = rsu.time_zero()
    info['time']['real']  = stop-start

    for watcher in reversed(watchers):
        watcher.stop ()
        watcher.join ()
        ru.dict_merge (info, watcher.get_data())

    # allow watchers to finalize some stuff, now having data from other watchers
    # available
    for watcher in reversed(watchers):
        watcher.finalize(info)

    time_2 = rsu.timestamp()
    load_2 = float(os.popen (LOAD_CMD).read())
    info['cpu']['load'] = max(load_1, load_2)
    rsu.logger.info ("system load %s: %s" % (LOAD, info['cpu']['load']))
   
    if  LOAD > 0:
        rsu.logger.info ("stopping system load")
        os.popen ("killall -9 stress 2>&1 > /dev/null")
        rsu.logger.info ("stopped  system load")

    return info, ret, out


_CPU = 'cpu'
_MEM = 'mem'
_STO = 'sto'
_TIM = 'time'
_NET = 'net'

_TYPE = 0
_TIME = 1
_VALS = 2

# ------------------------------------------------------------------------------
#
def _emulator(samples):

    atoms = dict()  # one atom of eeach type
    state = dict()  # there is at most one atom for each type in 'state'

    # create atoms for all sample types
    atoms[_TIM] = rsa.Time    ()
    atoms[_CPU] = rsa.Compute ()
    atoms[_MEM] = rsa.Memory  ()
    atoms[_STO] = rsa.Storage ()
    atoms[_NET] = rsa.Network ()

    # FIXME: make sure threads and queues are up
    time.sleep (0.1)

  # print "\n-------------------------\n\n"
    print "emulating %d samples" % len(samples)

    # run the first set of samples until we meet a sample type which is already
    # started.  At that point, start to wait before submission.  If all samples
    # have been run, wait for all atoms to complete, and voila

    idx = 0
    while idx < len(samples):

        t = samples[idx][_TYPE]
        v = samples[idx][_VALS]

        if t in state:
            # such an atom is running -- go into steady state to wait for it
          # print ' === brk %d : %-4s : %s' % (idx, t, v)
            break

        else:
            # no such atom running - start one
          # print ' === idx %d : %-4s : %s' % (idx, t, v)
            atoms[t].emulate(v)
            state[t] = atoms[t]

        idx += 1



    # we need to wait first before running the next sample of any type
    while idx < len(samples):

        t = samples[idx][_TYPE]
        v = samples[idx][_VALS]

        if t in state:
          # print ' === wai %d : %-4s : %s' % (idx, t, v)
            state[t].wait()
        else:
          # print ' === new %d : %-4s : %s' % (idx, t, v)
            state[t] = atoms[t]

      # print ' === idx %d : %s' % (idx, t)
        state[t].emulate(v)
        idx += 1

    # all samples are running now (or have been running), now wait for all
    # active ones
    for t in state:
      # print 'wai %s' % t
        state[t].wait()


    # we are done and can shut the atoms down
    for t in atoms:
        atoms[t].stop()

  # print "\n-------------------------\n\n"


# ------------------------------------------------------------------------------
#
def emulate(command=None, samples=None, src=None):

    if (command and samples) or \
       (command and src    ) or \
       (samples and src    )    :
        raise ValueError ("emulate needs *either* command, sample *or* src")

    if not command and not samples and not src:
        print "warning: emulate needs either command, sample or src"
        return[0, None, None]

    if command or src:

        if command:
            profs = rsu.get_profiles (command, mode='pro')

            # FIXME: average vals over all retrieved profiles
            prof  = profs[0]['profile']

        else:
            prof = ru.read_json(src)

      # pprint.pprint (prof)

        # get time series to emulate (all types of operations are mixed)
        # FIXME: we should also sample walltime for _TIM.  As it is, mixing
        #        time and other samples will yield incorrect results due to
        #        mismatch in granularity.
        # FIXME: add network sample interpretation
        samples  = list()
      # samples += [[_TIM, x[0], {'real'       : x[1].get('real',       0)}]
      #                           for x in prof['time']]

        samples += [[_CPU, x[0], {'time'       : x[1].get('time',       0),
                                  'flops'      : x[1].get('ops',        0),
                                  'efficiency' : x[1].get('efficiency', 0)}]
                                  for x in prof['cpu']['sequence']]

        samples += [[_MEM, x[0], {'size'       : x[1].get('size',       0)}]
                                  for x in prof['mem']['sequence']]

        samples += [[_STO, x[0], {'src'        : x[1].get('src',        None),
                                  'rsize'      : x[1].get('read',       0),
                                  'tgt'        : x[1].get('tgt',        None),
                                  'wsize'      : x[1].get('write',      0)}]
                                  for x in prof['sto']['sequence']]

    # *globally* sort samples by time
    samples = sorted (samples, key=lambda x: x[1])

  # print "samples:\n---"
  # pprint.pprint (samples)
  # print "---"

    watchmode = os.environ.get ('RADICAL_SYNAPSE_WATCHMODE')
    if not watchmode or watchmode.lower in ['none', 'noop']:
        start = time.time()
        _emulator(samples)
        stop  = time.time()

        ret   = None
        info  = dict()
        info['cmd']  = command
        info['time'] = dict()
        info['time']['start'] = start
        info['time']['real']  = stop-start

    else:
        # let the profiler know that we run an emulation, so that the profile is not
        # stored as 'application run'.
        os.environ['_RADICAL_SYNAPSE_EMULATED'] = 'TRUE'
        os.environ['_RADICAL_SYNAPSE_EMULATEE'] = command

        info, ret, _ = profile(_emulator, samples)

        if 'ops' in info['cpu']:
            info['cpu']['efficiency'] = info['cpu']['ops']                       \
                                        / ( info['cpu']['ops']                   \
                                          + info['cpu']['cycles_stalled_front']  \
                                          + info['cpu']['cycles_stalled_back']   \
                                          )

       #print 'efficiency = %s / (%s + %s + %s) = %s' % (
       #          info['cpu']['ops'],
       #          info['cpu']['ops'],
       #          info['cpu']['cycles_stalled_front'],
       #          info['cpu']['cycles_stalled_back'],
       #          info['cpu']['efficiency'])


    return (info, ret, None)




# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


from   _atoms    import atom_compute_asm
from   _atoms    import atom_compute
from   base      import AtomBase
from   constants import COMPUTE

OVERHEAD = 20   # in %

# ------------------------------------------------------------------------------
#
class Compute (AtomBase):
    """
    This Compute Synapse emulates a compute workload, i.e. it consumes
    a specified number of floating point operations.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self): 

        AtomBase.__init__ (self, COMPUTE)


    # --------------------------------------------------------------------------
    #
    def _verify(self, vals): 

        assert ('flops' in vals or 'time' in vals)

        # FIXME: empirical tuning factor toward 1 MFLOP
        vals['flops'] = int(vals.get('flops', 0) * 0.6)
        vals['time']  = int(vals.get('time',  0)      )


    # --------------------------------------------------------------------------
    #
    def _emulate (self, vals): 

        try:
            # TODO: switch between flops and time emulation
          # print "atom_compute (%s)" % vals['flops']
            atom_compute_asm (vals['flops'], vals['time'])

        except Exception as e:
            print "com atom error: %s" % e


#-------------------------------------------------------------------------------




#!/usr/bin/env python


import os
import sys
import time
import numpy
import pprint        as pp
import radical.utils as ru
import subprocess    as sp

import radical.synapse       as rs
import radical.synapse.atoms as rsa

# import pudb 
# pudb.set_interrupt_handler ()


host     = os.getenv ('HOST', os.popen ('hostname | cut -f 1 -d . | xargs echo -n').read ())
iters    = 10
mult_mem = 1
mult_io  = 10

# ------------------------------------------------------------------------------
#
def mandel (x=1024, y=1024, z=100) :

  # print 'mandel: %s %s %s' % (x, y, z)
    start = time.time ()

    x_pixels  = x
    y_pixels  = y
    iter_max  = z
    x_min     = -2.0
    x_max     = +1.0
    y_min     = -1.5
    y_max     = +1.5
    threshold = 1
    images    = list()

    for l in range (mult_mem) :
        images.append (list())

    for x in range (x_pixels) :

      # print x

        cx = x * (x_max - x_min) / x_pixels + x_min

        rgb_row = []

        for y in range (y_pixels):

            cy = y * (y_max - y_min) / (y_pixels - 1) + x_min

            c = complex (cx, cy)
            z = 0

            for i in range (iter_max) :
                
                if abs (z) > threshold: 
                    break 

                z = z * z + c 

            rgb_row.append (["%3d" % i, "%3d" % i, "%3d" % i])

        for l in range (mult_mem) :
            images[l].append (rgb_row)

    with open ("/tmp/mb.dat", "w") as f :
        for s in range (mult_io) :
            f.write (str(images[0]))


# ------------------------------------------------------------------------------
#
def synaptic (x, y, z, load_compute, load_memory, load_storage) :

    load_instances = 1

  # load_id = 'EMU.%04d' % x
  # print 'synaptic: %s %s %s' % (x, y, z)
  # print      '%8s: %s %s %s' % (load_id, load_compute, load_memory, load_storage)

    start = time.time()

    # create containers for different system workload types
    atoms = dict()
    atoms['c'] = rsa.Compute ()
    atoms['m'] = rsa.Memory  ()
    atoms['s'] = rsa.Storage ()

    # the atoms below are executed concurrently (in their own threads)
    atoms['c'].run (info={'n'   : load_compute})
    atoms['m'].run (info={'n'   : load_memory})
    atoms['s'].run (info={'n'   : load_storage,
                          'tgt' : '%(tmp)s/synapse_storage.tmp.%(pid)s'})

    # wait for all atom threads to be done
    info_c = atoms['c'].wait ()
    info_m = atoms['m'].wait ()
    info_s = atoms['s'].wait ()

    for info in [info_c, info_m, info_s] :
        for line in info['out'] :
            l = ru.ReString (line)
            if  l // '^(ru.\S+)\s+:\s+(\S+)$' :
                info[l.get()[0]] = l.get()[1]

    return {'c':info_c, 'm':info_m, 's':info_s}

# ------------------------------------------------------------------------------
#

n2h = rs.number_to_human

# host        load_id     t_real --------- --------- ---------     ?      load util  eff    load load_memory load_storage     x     y     z
header   = '%-10s %15s %5s %5s %5s stress   t_real     t_cpu     t_mem      t_io %9s %4s %4s %5s %9s %9s' % \
           ('# host', 'id', 'x', 'y', 'z', 'cpu', 'util', 'eff', 'load', 'mem', 'io')
print header

for stress in range(12,13) :
  # repeat first experiment, for warmup
  # for x in [10, 10, 20, 40, 80, 160, 320, 640, 1280, 2560, 5120, 10240] :
    for x in [2560] :
        for z in [250] : 

            y = x

            if  stress :
                s = str (stress)
                stresser = sp.Popen (["stress", "-q", "-c", s],
                                     stdout=None, stderr=None)

                time.sleep (5) # warmup time
    
    
            # --------------------------------------------------------------------------------------------
    
            iter_0_time = list()
    
            for iter in range (iters) :
    
                start = time.time ()
                mandel (x, y, z)
                stop  = time.time ()
    
                iter_0_time.append (stop - start)

                load_id  = 'EXE.%04d' % x
                stat_exe   = '%-10s %15s %5s %5s %5s %5d %9.2f' % \
                               (host, load_id, x, y, z, stress, stop-start)
                print stat_exe
    
            iter_0_mean_time = numpy.mean (numpy.array (iter_0_time))
            iter_0_std_time  = numpy.std  (numpy.array (iter_0_time))
    
            load_id  = 'STAT_EXE.%04d' % x
            stat_0   = '%-10s %15s %5s %5s %5s %5d %9.2f %9.2f' % \
                           (host, load_id, x, y, z, stress, 
                            iter_0_mean_time, iter_0_std_time)
            print stat_0

          # sp.call ("killall    stress", shell=True)
          # sp.call ("killall -9 stress", shell=True)
          #
          # continue
    
    
            # --------------------------------------------------------------------------------------------
            iter_1_time = list()
            iter_1_cpu  = list()
            iter_1_mem  = list()
            iter_1_io   = list()
            iter_1_util = list()
            iter_1_eff  = list()
            iter_1_sys  = list()
    
            for iter in range (iters) :
    
                info_pro, ret, out = rs.profile (mandel, x, y, z)
    
                load_compute = float(info_pro['cpu']['ops' ]) / (1024*1024) / 8
                load_memory  = float(info_pro['mem']['peak']) / (1024*1024)
                load_storage = float(info_pro['io']['write']) / (1024*1024)
    
                load_id  = 'PRO.%04d' % x
                output   = '%-10s %15s %5d %5d %5d %5d %9.2f --------- --------- --------- %9.2f %0.2f %0.2f %5.2f %9.2f %9.2f' % \
                           (host, load_id, x, y, z, stress,
                            float(info_pro['time']['real']), 
                            load_compute, 
                            info_pro['cpu']['utilization'], 
                            info_pro['cpu']['efficiency'], 
                            info_pro['sys']['load'],
                            load_memory, 
                            load_storage
                           )
                print output
                             
                iter_1_time .append(float(info_pro['time']['real']))
                iter_1_cpu  .append(float(load_compute))
                iter_1_mem  .append(float(load_memory))
                iter_1_io   .append(float(load_storage))
                iter_1_util .append(float(info_pro['cpu']['utilization']))
                iter_1_eff  .append(float(info_pro['cpu']['efficiency']))
                iter_1_sys  .append(float(info_pro['sys']['load']))
    
            iter_1_mean_time = numpy.mean (numpy.array (iter_1_time))
            iter_1_mean_cpu  = numpy.mean (numpy.array (iter_1_cpu ))
            iter_1_mean_mem  = numpy.mean (numpy.array (iter_1_mem ))
            iter_1_mean_io   = numpy.mean (numpy.array (iter_1_io  ))
            iter_1_mean_util = numpy.mean (numpy.array (iter_1_util))
            iter_1_mean_eff  = numpy.mean (numpy.array (iter_1_eff ))
            iter_1_mean_sys  = numpy.mean (numpy.array (iter_1_sys ))
    
            iter_1_std_time  = numpy.std  (numpy.array (iter_1_time))
            iter_1_std_cpu   = numpy.std  (numpy.array (iter_1_cpu ))
            iter_1_std_mem   = numpy.std  (numpy.array (iter_1_mem ))
            iter_1_std_io    = numpy.std  (numpy.array (iter_1_io  ))
            iter_1_std_util  = numpy.std  (numpy.array (iter_1_util))
            iter_1_std_eff   = numpy.std  (numpy.array (iter_1_eff ))
            iter_1_std_sys   = numpy.std  (numpy.array (iter_1_sys ))
            
            load_id  = 'MEAN_PRO.%04d' % x
            mean     = '%-10s %15s %5d %5d %5d %5d %9.2f --------- --------- --------- %9.2f %.2f %.2f %5.2f %9.2f %9.2f' % \
                       (host, load_id, x, y, z, stress,
                        iter_1_mean_time, 
                        iter_1_mean_cpu,
                        iter_1_mean_util, 
                        iter_1_mean_eff, 
                        iter_1_mean_sys,
                        iter_1_mean_mem, 
                        iter_1_mean_io
                       )
            print mean  
    
            load_id  = 'STD_PRO.%04d' % x
            std      = '%-10s %15s %5d %5d %5d %5d %9.2f --------- --------- --------- %9.2f %.2f %.2f %5.2f %9.2f %9.2f' % \
                       (host, load_id, x, y, z, stress, 
                        iter_1_std_time, 
                        iter_1_std_cpu,
                        iter_1_std_util, 
                        iter_1_std_eff, 
                        iter_1_std_sys,
                        iter_1_std_mem, 
                        iter_1_std_io
                       )
            print std
    
            load_id  = 'STAT_PRO.%04d' % x
            stat     = '%-10s %15s %5d %5d %5d %5d %9.2f %9.2f --------- --------- --------- --------- --------- --------- %9.2f %9.2f %.2f %.2f %.2f %.2f %5.2f %5.2f %9.2f %9.2f %9.2f %9.2f' % \
                       (host, load_id, x, y, z, stress, 
                        iter_1_mean_time, 
                        iter_1_std_time, 
                        iter_1_mean_cpu,
                        iter_1_std_cpu,
                        iter_1_mean_util, 
                        iter_1_std_util, 
                        iter_1_mean_eff, 
                        iter_1_std_eff, 
                        iter_1_mean_sys,
                        iter_1_std_sys,
                        iter_1_mean_mem, 
                        iter_1_std_mem, 
                        iter_1_mean_io,
                        iter_1_std_io
                       )
            print stat  
    
            # --------------------------------------------------------------------------------------------
           
            iter_2_time = list()
            iter_2_tcpu = list()
            iter_2_tmem = list()
            iter_2_tio  = list()
            iter_2_cpu  = list()
            iter_2_mem  = list()
            iter_2_io   = list()
            iter_2_util = list()
            iter_2_eff  = list()
            iter_2_sys  = list()

            iter_1_mean_cpu = 92881.19
            iter_1_mean_mem =  2016.00
            iter_1_mean_io  =  1437.55
    
            for iter in range (iters) :
    
                info_s, ret, out = rs.profile (synaptic, x, y, z,
                        iter_1_mean_cpu, iter_1_mean_mem, iter_1_mean_io)
    
              # print "#####################"
              # pp.pprint (ret['m'])
              # print "#####################"
    
                load_compute = float(info_s['cpu']['ops' ]) / (1024*1024) / 8
                load_memory  = float(ret['m']['ru.maxrss']) / (1024*1024)
                load_storage = float(info_s['io']['write']) / (1024*1024)
           
                load_id  = 'EMU.%04d' % x
                output   = '%-10s %15s %5d %5d %5d %5d %9.2f %9.2f %9.2f %9.2f %9.2f %0.2f %0.2f %5.2f %9.2f %9.2f' % \
                           (host, load_id, x, y, z, stress, 
                            float(info_s['time']['real']), 
                            ret['c']['timer'], 
                            ret['m']['timer'], 
                            ret['s']['timer'],
                            load_compute, 
                            info_s['cpu']['utilization'], 
                            info_s['cpu']['efficiency'], 
                            info_s['sys']['load'],
                            load_memory, 
                            load_storage
                            )
                print output
    
                iter_2_time .append(float(info_s['time']['real']))
                iter_2_tcpu .append(float(ret['c']['timer']))
                iter_2_tmem .append(float(ret['m']['timer']))
                iter_2_tio  .append(float(ret['s']['timer']))
                iter_2_cpu  .append(float(load_compute))
                iter_2_mem  .append(float(load_memory))
                iter_2_io   .append(float(load_storage))
                iter_2_util .append(float(info_s['cpu']['utilization']))
                iter_2_eff  .append(float(info_s['cpu']['efficiency']))
                iter_2_sys  .append(float(info_s['sys']['load']))
    
            iter_2_mean_time = numpy.mean (numpy.array (iter_2_time))
            iter_2_mean_tcpu = numpy.mean (numpy.array (iter_2_tcpu))
            iter_2_mean_tmem = numpy.mean (numpy.array (iter_2_tmem))
            iter_2_mean_tio  = numpy.mean (numpy.array (iter_2_tio ))
            iter_2_mean_cpu  = numpy.mean (numpy.array (iter_2_cpu ))
            iter_2_mean_mem  = numpy.mean (numpy.array (iter_2_mem ))
            iter_2_mean_io   = numpy.mean (numpy.array (iter_2_io  ))
            iter_2_mean_util = numpy.mean (numpy.array (iter_2_util))
            iter_2_mean_eff  = numpy.mean (numpy.array (iter_2_eff ))
            iter_2_mean_sys  = numpy.mean (numpy.array (iter_2_sys ))
    
            iter_2_std_time  = numpy.std  (numpy.array (iter_2_time))
            iter_2_std_tcpu  = numpy.std  (numpy.array (iter_2_tcpu))
            iter_2_std_tmem  = numpy.std  (numpy.array (iter_2_tmem))
            iter_2_std_tio   = numpy.std  (numpy.array (iter_2_tio ))
            iter_2_std_cpu   = numpy.std  (numpy.array (iter_2_cpu ))
            iter_2_std_mem   = numpy.std  (numpy.array (iter_2_mem ))
            iter_2_std_io    = numpy.std  (numpy.array (iter_2_io  ))
            iter_2_std_util  = numpy.std  (numpy.array (iter_2_util))
            iter_2_std_eff   = numpy.std  (numpy.array (iter_2_eff ))
            iter_2_std_sys   = numpy.std  (numpy.array (iter_2_sys ))
            
            load_id  = 'MEAN_EMU.%04d' % x
            mean     = '%-10s %15s %5d %5d %5d %5d %9.2f %9.2f %9.2f %9.2f %9.2f %.2f %.2f %5.2f %9.2f %9.2f' % \
                       (host, load_id, x, y, z, stress, 
                        iter_2_mean_time, 
                        iter_2_mean_tcpu, 
                        iter_2_mean_tmem,
                        iter_2_mean_tio, 
                        iter_2_mean_cpu,
                        iter_2_mean_util, 
                        iter_2_mean_eff, 
                        iter_2_mean_sys,
                        iter_2_mean_mem, 
                        iter_2_mean_io
                       )
            print mean  
    
            load_id  = 'STD_EMU.%04d' % x
            std      = '%-10s %15s %5d %5d %5d %5d %9.2f %9.2f %9.2f %9.2f %9.2f %.2f %.2f %5.2f %9.2f %9.2f' % \
                       (host, load_id, x, y, z, stress, 
                        iter_2_std_time, 
                        iter_2_std_tcpu,
                        iter_2_std_tmem,
                        iter_2_std_tio,
                        iter_2_std_cpu,
                        iter_2_std_util, 
                        iter_2_std_eff, 
                        iter_2_std_sys,
                        iter_2_std_mem, 
                        iter_2_std_io
                       )
            print std
    
            load_id  = 'STAT_EMU.%04d' % x
            stat     = '%-10s %15s %5d %5d %5d %5d %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %.2f %.2f %.2f %.2f %5.2f %5.2f %9.2f %9.2f %9.2f %9.2f' % \
                       (host, load_id, x, y, z, stress, 
                        iter_2_mean_time, 
                        iter_2_std_time, 
                        iter_2_mean_tcpu,
                        iter_2_std_tcpu,
                        iter_2_mean_tmem,
                        iter_2_std_tmem,
                        iter_2_mean_tio,
                        iter_2_std_tio,
                        iter_2_mean_cpu,
                        iter_2_std_cpu,
                        iter_2_mean_util, 
                        iter_2_std_util, 
                        iter_2_mean_eff, 
                        iter_2_std_eff, 
                        iter_2_mean_sys,
                        iter_2_std_sys,
                        iter_2_mean_mem, 
                        iter_2_std_mem, 
                        iter_2_mean_io,
                        iter_2_std_io
                       )
            print stat  
    
    
          # print ' ---------------------------------------------------------------'
          # print " MB  %-25s : %15.2f s" % ('time.real',               float(info_pro['time']['real'         ]))
          # print " MB  %-25s : %15.2f"   % ('sys.load',                float(info_pro['sys']['load'          ]))
          # print " MB  %-25s : %s"       % ('cpu.ops',            n2h (float(info_pro['cpu']['ops'           ]), rs.PREFIX_ISO, 'FLOP'  , "%(val)15.2f %(unit)s"))
          # print " MB  %-25s : %s"       % ('cpu.utilization',    n2h (float(info_pro['cpu']['utilization'   ]), rs.PREFIX_ISO, ''      , "%(val)15.2f %(unit)s"))
          # print " MB  %-25s : %s"       % ('cpu.efficiency',     n2h (float(info_pro['cpu']['efficiency'    ]), rs.PREFIX_ISO, ''      , "%(val)15.2f %(unit)s"))
          # print " MB  %-25s : %s"       % ('cpu.flops_per_core', n2h (float(info_pro['cpu']['flops_per_core']), rs.PREFIX_ISO, 'FLOP/s', "%(val)15.2f %(unit)s"))
          # print " MB  %-25s : %s"       % ('io.write',           n2h (float(info_pro['io']['write'          ]), rs.PREFIX_BIN, 'Byte'  , "%(val)15.2f %(unit)s"))
          # print " MB  %-25s : %s"       % ('mem.peak',           n2h (float(info_pro['mem']['peak'          ]), rs.PREFIX_BIN, 'Byte'  , "%(val)15.2f %(unit)s"))
          # print ' ---------------------------------------------------------------'                                                        
          # print " SYN %-25s : %15.2f s" % ('time.real',               float(info_s['time']['real'         ]))                            
          # print " SYN %-25s : %15.2f"   % ('sys.load',                float(info_s['sys']['load'          ]))
          # print " SYN %-25s : %s"       % ('cpu.ops',            n2h (float(info_s['cpu']['ops'           ]), rs.PREFIX_ISO, 'FLOP'  , "%(val)15.2f %(unit)s"))
          # print " SYN %-25s : %s"       % ('cpu.utilization',    n2h (float(info_s['cpu']['utilization'   ]), rs.PREFIX_ISO, ''      , "%(val)15.2f %(unit)s"))
          # print " SYN %-25s : %s"       % ('cpu.efficiency',     n2h (float(info_s['cpu']['efficiency'    ]), rs.PREFIX_ISO, ''      , "%(val)15.2f %(unit)s"))
          # print " SYN %-25s : %s"       % ('cpu.flops_per_core', n2h (float(info_s['cpu']['flops_per_core']), rs.PREFIX_ISO, 'FLOP/s', "%(val)15.2f %(unit)s"))
          # print " SYN %-25s : %s"       % ('io.write',           n2h (float(info_s['io']['write'          ]), rs.PREFIX_BIN, 'Byte'  , "%(val)15.2f %(unit)s"))
          # print " SYN %-25s : %s"       % ('mem.peak',           n2h (float(info_s['mem']['peak'          ]), rs.PREFIX_BIN, 'Byte'  , "%(val)15.2f %(unit)s"))
          # print ' ---------------------------------------------------------------'
          # pp.pprint (info_pro)                                  
          # print ' ---------------------------------------------------------------'
          # pp.pprint (info_s)                                  
          # print ' ---------------------------------------------------------------'
          # pp.pprint (ret)                                  
          # print ' ---------------------------------------------------------------'
    
            print

            # --------------------------------------------------------------------------------------------

            if  stress :
                stresser.terminate ()
                time.sleep (1)
                stresser.kill ()

                sp.call ("killall    stress", shell=True)
                sp.call ("killall -9 stress", shell=True)






__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


import os
import time
import errno
import psutil
import threading
import subprocess
import Queue

import radical.utils              as ru
import radical.utils.logger       as rul


# ------------------------------------------------------------------------------
#
class AtomBase (object):

    # --------------------------------------------------------------------------
    #
    def __init__  (self, atype):

        self._atype = atype
        self._pid   = os.getpid ()
        self._uid   = ru.generate_id ("%s" % self._atype)
        self.logger = ru.get_logger('radical.synapse.self._uid')

        # storage for temporary data and statistics
     #  self._tmpdir = "/scratch/synapse/" # FIXME
        self._tmpdir = "/tmp/"             # FIXME

        try:
            os.makedirs (self._tmpdir)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir (self._tmpdir):
                pass
            else: raise


        # start worker process
        self._work_queue   = Queue.Queue ()
        self._result_queue = Queue.Queue ()

        self._term  = threading.Event()
        self._proc  = threading.Thread (target=self.run)
        self._proc.start ()


    # --------------------------------------------------------------------------
    #
    def __str__  (self):

        return self._uid


    # --------------------------------------------------------------------------
    #
    def run (self):

        try:

            while not self._term.is_set():

                vals = self._work_queue.get()
              # print " === vals: %s" % vals

                if not vals:
                    # signal to finish
                    self._result_queue.put (True)
                    return

              # print "emulate %-20s: %s" % (self, vals)
                self._emulate(vals)
                self._result_queue.put (True)


        except Exception as e:
            print "main loop error in atom driver: %s" % e
            raise


    # --------------------------------------------------------------------------
    #
    def emulate(self, vals):

        try:
            self._verify(vals)
            print "emulate  %s" % vals
            self._work_queue.put(vals)

        except Exception as e:
            print 'emulation error: invalid data: %s' % vals
            ru.cancel_main_thread()


    # --------------------------------------------------------------------------
    #
    def wait (self):

        return self._result_queue.get ()


    # --------------------------------------------------------------------------
    #
    def stop (self):

        self._term.set()
        self._work_queue.put (None) # signal finish

        if  self._proc:
            self._proc.join ()


    # --------------------------------------------------------------------------
    #
    def _verify(self, vals):

        raise NotImplementedError('%s atom misses _verify()' % self._atype)


    # --------------------------------------------------------------------------
    #
    def _emulate(self, vals):

        raise NotImplementedError('%s atom misses _emulate()' % self._atype)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


import radical.utils as ru


from   _atoms    import atom_storage
from   base      import AtomBase
from   constants import STORAGE

# ------------------------------------------------------------------------------
#
class Storage (AtomBase):
    """
    This Storage Synapse emulates a storage workload, i.e. it allocates
    a specified storage size on disk.  It creates only minimal I/O to the
    allocated storage, to ensure that the full amount has been allocated.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self): 

        AtomBase.__init__ (self, STORAGE)


    # --------------------------------------------------------------------------
    #
    def _verify(self, vals): 

        if not vals.get('src'):
            vals['src'] = "/tmp/synapse_storage.in"

        if not vals.get('tgt'):
            vals['tgt'] = "/tmp/synapse_storage.%(pid)s.out"

        if not vals.get('buf'):
            vals['buf'] = 1024  # 1 k default buf size

        assert( ('rsize' in vals and 'src' in vals) or
                ('wsize' in vals and 'tgt' in vals)  )


    # --------------------------------------------------------------------------
    #
    def _emulate (self, vals):

        src   =     vals['src'  ] % { 'tmp' : self._tmpdir, 'pid' : self._pid   }
        rsize = int(vals['rsize'])
        tgt   =     vals['tgt'  ] % { 'tmp' : self._tmpdir, 'pid' : self._pid   }
        wsize = int(vals['wsize'])
        buf   = int(vals['buf'  ])

        try:
            atom_storage(src, rsize, tgt, wsize, buf)

        except Exception as e:
            print "sto atom error: %s" % e
            ru.cancel_main_thread()


#-------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2015, RADICAL@Rutgers"
__license__   = "MIT"


import os
import socket

import watcher_base as wb

from ..utils import human_to_number, PREFIX_ISO


# ------------------------------------------------------------------------------
#
class WatcherSys (wb.WatcherBase) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self, pid):

        wb.WatcherBase.__init__(self, pid)


    # --------------------------------------------------------------------------
    #
    def _pre_process (self, config): 

        self._data['sys'] = dict()
        self._data['cpu'] = dict()


    # --------------------------------------------------------------------------
    #
    def _post_process (self): 

        with open ("/proc/cpuinfo") as proc_cpuinfo:

            cpu_freq         = 1 # in Hz
            num_sockets      = 1
            cores_per_socket = 1 
            core_siblings    = 1
            threads_per_core = 1
            flops_per_cycle  = 4 # see wikipedia on FLOPS
            flops_per_core   = flops_per_cycle * cpu_freq

            for line in proc_cpuinfo.readlines ()  :

                if  line.startswith ('model name') :
                    elems = line.split ('@')
                    if  elems[-1].endswith ('Hz\n') :
                        cpu_freq = max(cpu_freq, human_to_number (elems[-1], prefix=PREFIX_ISO))

                if  line.startswith ('cpu MHz') :
                    elems = line.split (':')
                    cpu_freq = max(cpu_freq, float(elems[-1]) * 1000*1000)

                if  line.startswith ('physical id') :
                    elems = line.split (':')
                    num_sockets = max(num_sockets, int(elems[-1])+1)

                if  line.startswith ('cpu cores') :
                    elems = line.split (':')
                    cores_per_socket = max(cores_per_socket, int(elems[-1]))

                if  line.startswith ('siblings') :
                    elems = line.split (':')
                    core_siblings = max(core_siblings, int(elems[-1]))

            threads_per_core = int(core_siblings / cores_per_socket)
            flops_per_core   = int(cpu_freq * flops_per_cycle)

        self._data['cpu']['num_sockets'      ] = num_sockets      
        self._data['cpu']['cores_per_socket' ] = cores_per_socket 
        self._data['cpu']['threads_per_core' ] = threads_per_core 
        self._data['cpu']['frequency'        ] = cpu_freq         
        self._data['cpu']['flops_per_cycle'  ] = flops_per_cycle  
        self._data['cpu']['flops_per_core'   ] = flops_per_core   

        self._data['sys']['hostname'] = os.environ.get('RADICAL_SYNAPSE_HOSTNAME', 
                                                       socket.gethostname())
        



# ------------------------------------------------------------------------------






__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"



# ------------------------------------------------------------------------------
#
# Atom Types
#
UNKNOWN = 'unknown'
TIME    = 'time'
COMPUTE = 'compute'
COMPUTE = 'compute_asm'
MEMORY  = 'memory' 
STORAGE = 'storage'
NETWORK = 'network'





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2015, RADICAL@Rutgers"
__license__   = "MIT"


import watcher_base as wb

from ..utils import human_to_number


# ------------------------------------------------------------------------------
#
class WatcherSto (wb.WatcherBase) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self, pid):

        self._old_sample = dict()
        self._tot_sample = dict()

        wb.WatcherBase.__init__(self, pid)


    # --------------------------------------------------------------------------
    #
    def _pre_process  (self, config):

        self._f = open('/proc/%s/io' % self._pid, 'r')
        self._data['sto']             = dict()
        self._data['sto']['sequence'] = list()


    # --------------------------------------------------------------------------
    #
    def _post_process (self):

        self._f.close()

        # use the values from tot_sample as global total
        for key,val in self._tot_sample.iteritems():
            self._data['sto'][key] = val


    # --------------------------------------------------------------------------
    #
    def _sample (self, now) :

        try:
            self._f.seek(0,0)
            data = self._f.read()

        except Exception as e:
            # FIXME: use log
          # print "sto data source is gone (%s)" % e
            return

        sample = dict()
        keys   = {'rchar' : 'read',
                  'wchar' : 'write'}

        for line in data.split('\n'):
            if not ':' in line:
                continue
            key, val = line.split (':', 1)
            if  key.strip() in keys :
                sample[keys[key]] = human_to_number (val)


        # keep the max vals in tot_sample
        for key,val in sample.iteritems():
            self._tot_sample[key] = max(self._tot_sample.get(key, 0), val)


        # we don't want abs values but incremental changes
        for key in sample:

            val = sample[key]
            old = self._old_sample.get(key)

            if old:
                sample[key] = val - old

            self._old_sample[key] = val

        self._data['sto']['sequence'].append ([now, sample])


# ------------------------------------------------------------------------------






import os
import sys
import time
import glob
import signal
import pprint
import socket
import pymongo
import radical.utils        as ru

# import pudb
# pudb.set_interrupt_handler ()


LOAD = int (os.environ.get ('RADICAL_SYNAPSE_LOAD', '0'))

if 'darwin' not in sys.platform.lower():
  # LOAD_CMD    = "top -b -n1 | head -1  |       cut -f 4 -d :         | cut -f 1 -d ,"
    LOAD_CMD    = "top -b -n1 | head -n1 | rev | cut -f 3 -d \  | rev  | sed -e 's/,//'"
else:
    LOAD_CMD    = "top -l 1 | head -n 3 | tail -n 1 | cut -f 3 -d ' ' | cut -f 1 -d ','"


# ------------------------------------------------------------------------------
#
logger = ru.get_logger('radical.synapse')


# ------------------------------------------------------------------------------
#
_stamp_zero = None
def timestamp():
    global _stamp_zero
    if not _stamp_zero :
        _stamp_zero = time.time()
        return 0.0
    return time.time() - _stamp_zero

def time_zero():
    return _stamp_zero


# ------------------------------------------------------------------------------
#
#
PREFIX_BIN = { 'K' : 1024,
               'M' : 1024 * 1024,
               'G' : 1024 * 1024 * 1024,
               'T' : 1024 * 1024 * 1024 * 1024,
               'P' : 1024 * 1024 * 1024 * 1024 * 1024,
               'E' : 1024 * 1024 * 1024 * 1024 * 1024 * 1024,
               'Z' : 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024,
               'Y' : 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024
             }

PREFIX_ISO = { 'K' : 1000,
               'M' : 1000 * 1000,
               'G' : 1000 * 1000 * 1000,
               'T' : 1000 * 1000 * 1000 * 1000,
               'P' : 1000 * 1000 * 1000 * 1000 * 1000,
               'E' : 1000 * 1000 * 1000 * 1000 * 1000 * 1000,
               'Z' : 1000 * 1000 * 1000 * 1000 * 1000 * 1000 * 1000,
               'Y' : 1000 * 1000 * 1000 * 1000 * 1000 * 1000 * 1000 * 1000
             }


# ------------------------------------------------------------------------------
#
def human_to_number (h, prefix=PREFIX_BIN) :

    rs = ru.ReString (h)

    with rs // '^\s*([\d\.]+)\s*(\D+)\s*$' as match :
        if  not match :
          # print 'incorrect format: %s' % h
            return float(h)

        p = match[1].upper()[0]

        if  not p in prefix :
          # print 'unknown prefix: %s' % h
            return float(h)

        return float(match[0]) * prefix[p]


# ------------------------------------------------------------------------------
#
def number_to_human (n, prefix=PREFIX_BIN, unit='', template="%(val)f %(unit)s") :

    for key in prefix.keys () :

        hn = n / float(prefix[key])
        if  hn > 1 and hn < 1000 :
            return template % {'val' : hn, 'unit' : "%s%s" % (key, unit)}

    return template % {'val' : n, 'unit' : unit}


# ------------------------------------------------------------------------------
#
def time_to_seconds (t) :

    rs = ru.ReString (t)

    with rs // '^(?:\s*(\d+)\s*[:])+\s*([\d\.]+)\s*$' as match :

        if  not match :
            return t

        seconds = 0
        if  len(match) ==  1 : seconds = float(match[0])
        if  len(match) ==  2 : seconds = float(match[0]) * 60      + float(match[1])
        if  len(match) ==  3 : seconds = float(match[0]) * 60 * 60 + float(match[1]) * 60 + float(match[2])

        return seconds


# ------------------------------------------------------------------------------
def store_profile (profile, tags=None, url=None, mode=None) :

    if not url:
        url = os.environ.get ('RADICAL_SYNAPSE_DBURL')

    if not url:
      # print "warning: need dburl to store profiles"
        return None

    if not mode:
        raise ValueError ("document needs mode (emulated | eecuted | profiled)")

    url = ru.Url (url)

    if not tags:
        tags  = dict()
        elems = filter (None, os.environ.get('RADICAL_SYNAPSE_TAGS', '').split(','))
        for elem in elems:
            if ':' in elem:
                key, val  = elem.split(':', 1)
                tags[key] = val
            else:
                tags[elem] = None


    command_idx = index_command (profile['cmd'], tags)
    print "index %s (%s) to %s" % (profile['cmd'], tags, command_idx)

    host = profile['sys'].get ('hostname')
    if not host:
        host = os.environ.get ('RADICAL_SYNAPSE_HOSTNAME', socket.gethostname())
        profile['sys']['hostname'] = host

    doc  = {'type'        : 'synapse_profile',
            'mode'        : mode,
            'command_idx' : command_idx,
            'command'     : profile['cmd'],
            'tags'        : tags,
            'profile'     : profile}



    if url.schema == 'mongodb':

        print 'store profile in db %s' % url

        [dbhost, port, dbname, _, _, _, _] = ru.split_dburl (url)

        db_client  = pymongo.MongoClient (host=dbhost, port=port)
        database   = db_client[dbname]
        collection = database['profiles']

        collection.insert (doc)


    elif url.schema == 'file':

        path = url.path

        if not os.path.isdir (path):
            os.system ('mkdir -p "%s"' % path)

        name = command_idx.split()[0]
      # for key, val in tags.iteritems():
      #     if val != None: name += "_%s:%s" % (key, val)
      #     else          : name += "_%s"    % (key)
        for tag in sorted(tags.keys()):
            if tags[tag] != None: name += "_%s" % tags[tag]
            else                : name += "_%s" % tag

        idx  = 0
        while True:
            fname = "%s/synapse_profile_%s_%s_%s_%03d.json" % (path, name, host, mode[0:3], idx)
            if not os.path.exists (fname):
                break
            idx += 1

        print 'store profile in file %s' % fname
        os.system ('mkdir -p "%s/"' % path)
        ru.write_json (doc, fname)



# ------------------------------------------------------------------------------
def get_profiles (command, tags=None, url=None, mode=None) :

    print command

    if not url:
        url = os.environ.get ('RADICAL_SYNAPSE_DBURL')

    if not url:
        print "warning: need dburl to retrieve profiles"
        return None

    url = ru.Url(url)

    if mode and not isinstance (mode, list):
        mode = [mode]

    if not tags:
        tags  = dict()
        elems = filter (None, os.environ.get('RADICAL_SYNAPSE_TAGS', '').split(','))
        for elem in elems:
            if ':' in elem:
                key, val  = elem.split(':', 1)
                tags[key] = val
            else:
                tags[elem] = None


    command_idx = index_command (command, tags)

    if url.schema == 'mongodb':

        [dbhost, port, dbname, _, _, _, _] = ru.split_dburl (url)

        db_client  = pymongo.MongoClient (host=dbhost, port=port)
        database   = db_client[dbname]
        collection = database['profiles']

        # FIXME: eval partial tags

        if mode:
            results = collection.find ({'type'        : 'synapse_profile',
                                        'tags'        : tags,
                                        'mode'        : {'$in': mode},  # FIXME: check
                                        'command_idx' : command_idx})
        else:
            results = collection.find ({'type'        : 'synapse_profile',
                                        'tags'        : tags,
                                        'command_idx' : command_idx})

        if  not results.count() :
            raise RuntimeError ("Could not get profile for %s at %s/profiles"
                    % (command, url))

        ret = list(results)


    elif url.schema == 'file':

        path = url.path

        if not os.path.isdir (path):
            raise ValueError ("dburl (%s) must point to an existing dir" % url)

        name = command_idx.split()[0]
      # for key, val in tags.iteritems():
      #     if val != None: name += "_%s:%s" % (key, val)
      #     else          : name += "_%s"    % (key)
        for tag in sorted(tags.keys()):
            if tags[tag] != None: name += "_%s" % tags[tag]
            else                : name += "_%s" % tag

      # print    "checking profiles %s/synapse_profile_%s_*.json" % (path, name)
        base   = "%s/synapse_profile_%s_*.json" % (path, name)
        fnames = glob.glob (base)
        ret    = list()
        for fname in fnames:

          # print 'reading profile %s' % fname

            doc     = ru.read_json_str (fname)
            use     = False
            docmode = doc['mode'][0:3]

            doc['fname'] = fname

            if doc['command'] == command:
                if not mode :
                    use = True
                elif docmode in mode:
                    use = True
                else:
                    print "skip: mode %s not in %s" % (docmode, mode)
            else:
                print "skip command %s" % command
                print "   ! command %s" % doc['command']

            if use:
                ret.append (doc)

        if not len(ret):
            raise LookupError ("No matching profile at %s" % base)

  # print 'retrieved %d profiles from %s' % (len(ret), url)
  # pprint.pprint (ret)

    return ret

# ------------------------------------------------------------------------------
#
def get_all_frames (command, tags=None, url=None, mode=None) :

    if not url:
        url = os.environ.get ('RADICAL_SYNAPSE_DBURL')

    if not url:
        print "warning: need dburl to retrieve profiles"
        return None

    docs = list()

    url = ru.Url (url)
    if url.scheme != 'file':
        raise ValueError ('can only handle file:// based dburls, not %s' % url)

    path = url.path

    for p in glob.glob ("%s/*/" % path):
        tmp_url = ru.Url (url)
        tmp_url.path = p
        tmp_docs = get_profiles (command, tags=tags, url=tmp_url, mode=mode)
        if tmp_docs:
            docs += tmp_docs

    return make_frames (docs)


# ------------------------------------------------------------------------------
#
def get_frames (command, tags=None, url=None, mode=None) :

    if not url:
        url = os.environ.get ('RADICAL_SYNAPSE_DBURL')

    if not url:
        print "warning: need dburl to retrieve profiles"
        return None

    docs = get_profiles (command, tags, url, mode)

    return make_frames (docs)


# ------------------------------------------------------------------------------
#
def make_frames (docs):

    frame_dicts = list()

    for doc in docs:

        profile = doc['profile']

        tags = doc['tags']
        mode = doc['mode'][0:3]

        cmd  = profile.get ('cmd',  "")
        sys  = profile.get ('sys',  {})
        sto  = profile.get ('sto',  {})
        cpu  = profile.get ('cpu',  {})
        mem  = profile.get ('mem',  {})
        sys  = profile.get ('sys',  {})
        time = profile.get ('time', {})

        frame_dict = {
            'host'         : sys['hostname'],
            'mode'         : mode,
            'tags'         : tags,
            'cmd'          : doc['command'],
            'cmd_idx'      : doc['command_idx'],

            'time_real'    : time['real'],
            'time_start'   : time['start'],

            'cpu_ops'      : cpu.get('ops'),
            'cpu_flops'    : cpu.get('flops'),
            'cpu_sequence' : cpu.get('sequence'),

            'mem_peak'     : mem.get('peak'),
            'mem_rss'      : mem.get('rss'),
            'mem_sequence' : mem.get('sequence'),
            'mem_size'     : mem.get('size'),

            'sto_read'     : sto.get('read'),
            'sto_write'    : sto.get('write'),
            'sto_sequence' : sto.get('sequence'),
        }

        for tag in tags:
            # FIXME: assumption is that tags are ints -- that is not save
            frame_dict['tag_%s'%tag] = float(tags[tag])

        frame_dicts.append(frame_dict)

    import pandas 
    frames = pandas.DataFrame (frame_dicts)

    return frames


# ------------------------------------------------------------------------------
#
def index_command (command, tags=None) :

    return str(command)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "LGPL.v3"


import radical.utils as ru


from   _atoms    import atom_memory
from   base      import AtomBase
from   constants import MEMORY

# ------------------------------------------------------------------------------
#
class Memory (AtomBase):
    """
    This Memory Synapse emulates a memory workload, i.e. it allocates
    a specified memory size.  It does not create nor consume any I/O to the
    allocated memory -- instead, it quits right after allocation.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self): 

        AtomBase.__init__ (self, MEMORY)


    # --------------------------------------------------------------------------
    #
    def _verify(self, vals): 

        assert ('size' in vals)


    # --------------------------------------------------------------------------
    #
    def _emulate (self, vals): 

        try:
         # print "atom_memorye (%s)" % vals['size']
            atom_memory (int(vals['size']))

        except Exception as e:
            print "mem atom error: %s" % e
            ru.cancel_main_thread()



#-------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"



# ------------------------------------------------------------------------------

from synapsify import synapsify
from synapsify import NOTHING
from synapsify import PROFILE
from synapsify import EMULATE

from synapse   import profile
from synapse   import emulate
from synapse   import execute



# ------------------------------------------------------------------------------
#

import os
import radical.utils as ru

pwd     = os.path.dirname (__file__)
root    = "%s/.." % pwd
version, version_detail, version_branch, sdist_name, sdist_path = ru.get_version ([root, pwd])

_logger = ru.get_logger('radical.synapse')


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2015, RADICAL@Rutgers"
__license__   = "MIT"

import re
import os
import signal
import select
import threading

import subprocess    as sp
import radical.utils as ru
import watcher_base  as wb


# ------------------------------------------------------------------------------
#
class WatcherCPU (wb.WatcherBase) :
    """
    This watcher mostly uses 'perf stat' to analyse program execution
    performance.  See [1] for an detailed introduction to perf.
    
    [1] http://www.brendangregg.com/perf.html
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self, pid):

        wb.WatcherBase.__init__(self, pid)

        self._data['cpu']             = dict()
        self._data['cpu']['sequence'] = list()


    # --------------------------------------------------------------------------
    #
    # FIXME: for some reasons, running w/o shell=True fails.   So we need to
    # take a detour to communicate the perf pid.  We should try to get rid of
    # the shelling...
    #
    def _pre_process (self, config):

        # sample rate is in samples/sec, we convert into time between samples in
        # milliseconds.  Minimum is 100 ms
        sample_time = 1 / config.get ('sample_rate', 1) * 1000
        sample_time = max(sample_time, 100)

        sample_cmd  = "sh -c 'perf stat -I %d -p %d & PID=$!; " % (sample_time, self._pid) \
                    + "echo $PID > /tmp/synapse.pid.cpu_sample.$PPID; " \
                    + "wait $PID'" 
        total_cmd   = "sh -c 'perf stat -v -p %d & PID=$!; " % (self._pid) \
                    + "echo $PID > /tmp/synapse.pid.cpu_total.$PPID; " \
                    + "wait $PID'" 

      # print sample_cmd
      # print total_cmd
        self._psam  = sp.Popen (sample_cmd,
                                stdout = sp.PIPE,
                                stderr = sp.STDOUT,
                                shell  = True)
        self._ptot  = sp.Popen (total_cmd,
                                stdout = sp.PIPE,
                                stderr = sp.STDOUT,
                                shell  = True)

        self._osam  = ""
        self._otot  = ""


    # --------------------------------------------------------------------------
    #
    def _sample (self, now):

        # to avoid the proc buffers to fill up, we need to continuously read
        # from the pipes.  We don't want to have this blocking though, so only
        # read if select tells us there are more data

        while (select.select([self._psam.stdout], [], [], 0)[0] != []):   
            data        = self._psam.stdout.read(1)
            self._osam += data

        while (select.select([self._ptot.stdout], [], [], 0)[0] != []):   
            data        = self._ptot.stdout.read(1)
            self._otot += data

    # --------------------------------------------------------------------------
    #
    def _post_process (self):

        # proc should be done now -- let it know.  But first make sure we are
        # listening on the pipes when it dies...
        perf_pid = int(open ('/tmp/synapse.pid.cpu_total.%s' % self._ptot.pid, 'r').read().strip())
        os.unlink ('/tmp/synapse.pid.cpu_total.%s' % self._ptot.pid)

        threading.Timer (1.0, os.kill, [perf_pid, signal.SIGINT]).start ()
        out = self._ptot.communicate()[0]

        self._parse_perf_total (self._otot + out)


        # now do the same for the sampling counters
        perf_pid = int(open ('/tmp/synapse.pid.cpu_sample.%s' % self._psam.pid, 'r').read().strip())
        os.unlink ('/tmp/synapse.pid.cpu_sample.%s' % self._psam.pid)

        threading.Timer (1.0, os.kill, [perf_pid, signal.SIGINT]).start ()
        out = self._psam.communicate()[0]

        self._parse_perf_sample (self._osam + out)


    # --------------------------------------------------------------------------
    # 
    def _finalize (self, info) :

        # efficiency  = cycles_used /  cycles_spent
        #             = cycles_used / (cycles_used + cycles_wasted)
        #             = 0..1
        # utilization = cycles_used /  cycles_max
        #             = 0..1
        #
        # In 'perf stat', stalled cycles + used cycles don't add up to spent
        # cycles (ie. usually gove more than 100%).  This is due to the fact
        # that different low level counter layers are used which can be
        # triggered by front- and backend stalls.  The used definitions
        # for utilization and efficiency may thus not exactly reflect the
        # naively expected definitions -- but they should nevertheless
        # result in a useful measure with similar bouncs as the naive
        # definition...
        #
        #  * http://stackoverflow.com/questions/22165299/

        import pprint
        pprint.pprint(info['cpu'])

        opc  = info['cpu']['flops_per_cycle']
        fpc  = info['cpu']['flops_per_cycle']
        freq = info['cpu']['frequency']

        real = info['time']['real']

        ops  = info['cpu'].get('ops', 0)
        cns  = info['cpu'].get('cycles', 0)
        csf  = info['cpu'].get('cycles_stalled_front', 0)
        csb  = info['cpu'].get('cycles_stalled_back',  0)

        omax = cns * fpc

        if real    : info['cpu']['real']        = real
        else       : info['cpu']['real']        = None

        if real    : info['cpu']['flops']       = ops / real
        else       : info['cpu']['flops']       = None

        if opc     : info['cpu']['efficiency']  = fpc / opc
        else       : info['cpu']['efficiency']  = None

        if ops     : info['cpu']['utilization'] = omax / ops
        else       : info['cpu']['utilization'] = None
        

        # now we do the same for each sample, and use the time diff between
        # samples as real time.
        old = 0.0
        for ts,sample in info['cpu']['sequence']:

            real = ts - old
            old  = ts

            ops  = sample.get('ops', 0)
            cns  = sample.get('cycles', 0)
            csf  = sample.get('cycles_stalled_front', 0)
            csb  = sample.get('cycles_stalled_back',  0)
            opc  = sample.get('ops_per_cycle')

            omax = cns * fpc

            if real    : sample['real']        = real
            else       : sample['real']        = None

            if real    : sample['flops']       = ops / real
            else       : sample['flops']       = None

            if opc     : sample['efficiency']  = fpc / opc
            else       : sample['efficiency']  = None

            if ops     : sample['utilization'] = omax / ops
            else       : sample['utilization'] = None


    # --------------------------------------------------------------------------
    # 
    def _parse_perf_total (self, perf_out) :
    
        if  isinstance (perf_out, basestring) :
            perf_out = perf_out.split ('\n')
    
        # prepare to dig data from perf output lines
        perf_keys  = {# "CPUs utilized"           : "utilization",
                        "instructions"            : "ops",
                        "branches"                : "branches",
                        "branch-misses"           : "branch_misses",
                        "cycles"                  : "cycles",
                        "stalled-cycles-frontend" : "cycles_stalled_front",
                        "stalled-cycles-backend"  : "cycles_stalled_back",
                        "frontend cycles idle"    : "cycles_idle_front",
                        "backend  cycles idle"    : "cycles_idle_back",
                        "insns per cycle"         : "ops_per_cycle"}
        ored_keys   = '|'.join(perf_keys.keys()).replace (' ', '\s')
        perf_patstr = r"""
            ^(?P<lead>.*?\s+)            # lead-in
            (?P<val>[\d\.,]+)%%?         # value
            \s+                          # skip
            (?P<key>%s)                  # key
            \s*                          # skip
            (\[(?P<perc>[\d\.]+)%%\])?   # percentage (optional)
            \s*                          # skip
            (?P<rest>.*)$                # lead-out
        """ % ored_keys
        perf_pat = re.compile (perf_patstr, re.VERBOSE)
    
        # and go
        for line in perf_out :
    
            l = ru.ReString (line)
    
          # print "line: %s" % line
    
            while l // (perf_pat) :
    
                key  = perf_keys[l.get ('key')]
                val  =           l.get ('val')
                perc =           l.get ('perc', '-1.0')
    
                if  not perc :
                    perc = '-1.0'
    
              # print " ->  %s  %s  %s" % (key, val, perc)
    
                self._data['cpu']['%s'      % key] = float(val.replace  (',', ''))
                self._data['cpu']['%s_perc' % key] = float(perc.replace (',', ''))
    
              # print "rest: %s" % l.get ('rest')
                l = ru.ReString (l.get ('rest'))
    
        # must haves
        if not self._data['cpu'].get('ops'                 ) : self._data['cpu']['ops'                  ] = 1
        if not self._data['cpu'].get('cycles_idle_front'   ) : self._data['cpu']['cycles_idle_front'    ] = 0
        if not self._data['cpu'].get('cycles_idle_back'    ) : self._data['cpu']['cycles_idle_back'     ] = 0
        if not self._data['cpu'].get('cycles_stalled_front') : self._data['cpu']['cycles_stalled_front' ] = 0
        if not self._data['cpu'].get('cycles_stalled_back' ) : self._data['cpu']['cycles_stalled_back'  ] = 0
    
        self._data['cpu']['efficiency']  = self._data['cpu']['ops']                       \
                              / ( self._data['cpu']['ops']                   \
                                + self._data['cpu']['cycles_stalled_front']  \
                                + self._data['cpu']['cycles_stalled_back']   \
                                )
    
        # also determine the theoretical number of FLOPS, which is calculated as
        #   FLOPS = #cores * #cycles/sec * flops/cycle
        # where flops/cycle are assumed to be 4 (see wikipedia on FLOPS).  We assume
        # that the watched process uses only one core, so we calculate the number
        # for one core only, but also report the number of cores.
        
    
    
    # --------------------------------------------------------------------------
    # 
    def _parse_perf_sample (self, perf_out) :
    
        if  isinstance (perf_out, basestring) :
            perf_out = perf_out.split ('\n')
    
        # prepare to dig data from perf output lines
        perf_keys  = {# "task-clock"              : "utilization",
                      # "context-switches"        : "context_switches",
                      # "cpu-migrations"          : "cpu_migrations",
                        "instructions"            : "ops",
                      # "page-faults"             : "page_faults",
                        "branches"                : "branches",
                        "branch-misses"           : "branch_misses",
                        "cycles"                  : "cycles",
                        "stalled-cycles-frontend" : "cycles_stalled_front",
                        "stalled-cycles-backend"  : "cycles_stalled_back"}
        ored_keys   = '|'.join(perf_keys.keys()).replace (' ', '\s')
        perf_patstr = r"""
           ^(?P<lead>\s+)                # lead-in
            (?P<time>[\d\.,]+)           # timestamp
            \s+                          # skip
            (?P<val>[\d\.,]+)            # value (ignore warnings)
            \s+                          # skip
            (?P<key>%s)                  # key
            \s*                          # skip
            (\[(?P<perc>[\d\.]+)%%\])?   # percentage (optional)
            \s*                          # skip
            (?P<rest>.*)$                # lead-out
        """ % ored_keys
        perf_pat = re.compile (perf_patstr, re.VERBOSE)
    
        sample   = dict() 
        last_ts  = None
        ts       = None

        # and go
        for line in perf_out :
    
            l = ru.ReString (line)
    
          # print "line: %s" % line
    
            while l // (perf_pat) :
    
                ts   =     float(l.get ('time'))
                key  = perf_keys[l.get ('key')]
                val  =           l.get ('val')
                perc =           l.get ('perc')
    
                if  not perc :
                    perc = '-1.0'
    
              # print " ->  %s/%s  %s  %s  %s" % (ts, last_ts, key, val, perc)
    
                if ts != last_ts:
                    if last_ts != None:
                      # print "append %s" % last_ts
                        self._data['cpu']['sequence'].append ([last_ts, sample])
                        sample = dict()
                    last_ts = ts
    
                sample[key] = float(val.replace  (',', ''))
                if perc != None:
                    sample['%s_perc' % key] = float(perc.replace  (',', ''))
    
              # print "rest: %s" % l.get ('rest')
                l = ru.ReString (l.get ('rest'))
    
        if ts and sample:
          # print "append %s" % ts
            self._data['cpu']['sequence'].append ([ts, sample])
    
    
# ------------------------------------------------------------------------------
    



#!/usr/bin/env python


import os
import sys
import time
import numpy
import pprint        as pp
import radical.utils as ru

import radical.synapse       as rs
import radical.synapse.atoms as rsa

# import pudb 
# pudb.set_interrupt_handler ()

iters = 10
host  = os.getenv ('HOST', os.popen ('hostname | cut -f 1 -d . | xargs echo -n').read ())

# ------------------------------------------------------------------------------
#
def synaptic (load_compute, load_memory, load_storage) :

    start = time.time()

    # create containers for different system workload types
    atoms = dict()
    atoms['c'] = rsa.Compute ()
    atoms['m'] = rsa.Memory  ()
    atoms['s'] = rsa.Storage ()

    # the atoms below are executed concurrently (in their own threads)
    atoms['c'].run (info={'n'   : load_compute})
    atoms['m'].run (info={'n'   : load_memory})
    atoms['s'].run (info={'n'   : load_storage,
                          'tgt' : '%(tmp)s/synapse_storage.tmp.%(pid)s'})

    # wait for all atom threads to be done
    info_c = atoms['c'].wait ()
    info_m = atoms['m'].wait ()
    info_s = atoms['s'].wait ()

    for info in [info_c, info_m, info_s] :
        for line in info['out'] :
            l = ru.ReString (line)
            if  l // '^(ru.\S+)\s+:\s+(\S+)$' :
                info[l.get()[0]] = l.get()[1]

    return {'c':info_c, 'm':info_m, 's':info_s}


# ------------------------------------------------------------------------------
#

def main (cfg_list) :

    n2h = rs.number_to_human

    for cfg in cfg_list :
        x       = int  (cfg[ 2])
        y       = int  (cfg[ 3])
        z       = int  (cfg[ 4])
        stress  = int  (cfg[ 5])
        compute = float(cfg[10])
        util    = float(cfg[11])
        eff     = float(cfg[12])
        load    = float(cfg[13])
        memory  = float(cfg[14])
        storage = float(cfg[15])

        iter_1_time = list()
        iter_1_cpu  = list()
        iter_1_mem  = list()
        iter_1_io   = list()
        iter_1_util = list()
        iter_1_eff  = list()
        iter_1_sys  = list()

        for iter in range (iters) :

            info_emu, ret, out = rs.profile (synaptic, compute, memory, storage)

            if not 'utilization' in info_emu['cpu'] : info_emu['cpu']['utilization' ] = 0
            if not 'efficiency'  in info_emu['cpu'] : info_emu['cpu']['efficiency'  ] = 0
            if not 'load'        in info_emu['cpu'] : info_emu['cpu']['load'        ] = 0
            if not 'ops'         in info_emu['cpu'] : info_emu['cpu']['ops'         ] = 0
            if not 'max'         in info_emu['mem'] : info_emu['mem']['max'         ] = 0
            if not 'write'       in info_emu['io']  : info_emu['io']['write'        ] = 0
        
            load_compute = int(float(info_emu['cpu']['ops' ]) / (1024*1024) / 8)
            load_memory  = int(float(info_emu['mem']['max' ]) / (1024*1024))
            load_storage = int(float(info_emu['io']['write']) / (1024*1024))
            
            load_compute = int(float(info_emu['cpu']['ops' ]) / (1024*1024) / 8)
            load_memory  = int(float(info_emu['mem']['max' ]) / (1024*1024))
            load_storage = int(float(info_emu['io']['write']) / (1024*1024))
            
            load_id  = 'EMU.%04d' % x
            output   = '%-10s %15s %5d %5d %5d %6d %8.2f --------- --------- --------- %9.2f %0.2f %0.2f %5.2f %9.2f %9.2f' % \
                       (host, load_id, x, y, z, stress,
                        float(info_emu['time']['real']), 
                        load_compute, 
                        info_emu['cpu']['utilization'], 
                        info_emu['cpu']['efficiency'], 
                        info_emu['sys']['load'],
                        load_memory, 
                        load_storage
                       )
            print output

            iter_1_time .append(float(info_emu['time']['real']))
            iter_1_cpu  .append(float(load_compute))
            iter_1_mem  .append(float(load_memory))
            iter_1_io   .append(float(load_storage))
            iter_1_util .append(float(info_emu['cpu']['utilization']))
            iter_1_eff  .append(float(info_emu['cpu']['efficiency']))
            iter_1_sys  .append(float(info_emu['sys']['load']))

        iter_1_mean_time = numpy.mean (numpy.array (iter_1_time))
        iter_1_mean_cpu  = numpy.mean (numpy.array (iter_1_cpu ))
        iter_1_mean_mem  = numpy.mean (numpy.array (iter_1_mem ))
        iter_1_mean_io   = numpy.mean (numpy.array (iter_1_io  ))
        iter_1_mean_util = numpy.mean (numpy.array (iter_1_util))
        iter_1_mean_eff  = numpy.mean (numpy.array (iter_1_eff ))
        iter_1_mean_sys  = numpy.mean (numpy.array (iter_1_sys ))
    
        iter_1_std_time  = numpy.std  (numpy.array (iter_1_time))
        iter_1_std_cpu   = numpy.std  (numpy.array (iter_1_cpu ))
        iter_1_std_mem   = numpy.std  (numpy.array (iter_1_mem ))
        iter_1_std_io    = numpy.std  (numpy.array (iter_1_io  ))
        iter_1_std_util  = numpy.std  (numpy.array (iter_1_util))
        iter_1_std_eff   = numpy.std  (numpy.array (iter_1_eff ))
        iter_1_std_sys   = numpy.std  (numpy.array (iter_1_sys ))
        
        load_id  = 'MEAN_EMU.%04d' % x
        mean     = '%-10s %15s %5d %5d %5d %5d %9.2f --------- --------- --------- %9.2f %.2f %.2f %5.2f %9.2f %9.2f' % \
                   (host, load_id, x, y, z, stress,
                    iter_1_mean_time, 
                    iter_1_mean_cpu,
                    iter_1_mean_util, 
                    iter_1_mean_eff, 
                    iter_1_mean_sys,
                    iter_1_mean_mem, 
                    iter_1_mean_io
                   )
        print mean  
    
        load_id  = 'STD_EMU.%04d' % x
        std      = '%-10s %15s %5d %5d %5d %5d %9.2f --------- --------- --------- %9.2f %.2f %.2f %5.2f %9.2f %9.2f' % \
                   (host, load_id, x, y, z, stress, 
                    iter_1_std_time, 
                    iter_1_std_cpu,
                    iter_1_std_util, 
                    iter_1_std_eff, 
                    iter_1_std_sys,
                    iter_1_std_mem, 
                    iter_1_std_io
                   )
        print std
    
        load_id  = 'STAT_EMU.%04d' % x
        stat     = '%-10s %15s %5d %5d %5d %5d %9.2f %9.2f --------- --------- --------- --------- --------- --------- %9.2f %9.2f %.2f %.2f %.2f %.2f %5.2f %5.2f %9.2f %9.2f %9.2f %9.2f' % \
                   (host, load_id, x, y, z, stress, 
                    iter_1_mean_time, 
                    iter_1_std_time, 
                    iter_1_mean_cpu,
                    iter_1_std_cpu,
                    iter_1_mean_util, 
                    iter_1_std_util, 
                    iter_1_mean_eff, 
                    iter_1_std_eff, 
                    iter_1_mean_sys,
                    iter_1_std_sys,
                    iter_1_mean_mem, 
                    iter_1_std_mem, 
                    iter_1_mean_io,
                    iter_1_std_io
                   )
        print stat  
    
# ------------------------------------------------------------------------------
#
#
cfg_list = list ()

if  not len(sys.argv) > 1 :
    print "\n\tusage: %s <cfg_file> \n\n" % sys.argv[0]
    sys.exit (-1)

with open (sys.argv[1]) as fin :
    lines = fin.readlines ()

for line in lines :
    line = line.replace ('\n', '')
    re = ru.ReString (line)
    if  re // r'(MEAN_PRO)' :
        print line
        cfg_list.append (line.split())

main (cfg_list)

  





import radical.utils as ru

import atoms as rsa

# 
# 
# # ------------------------------------------------------------------------------
# class Description (object) :
# 
#     # --------------------------------------------------------------------------
#     def __init__ (self, fname) :
#         """
#         fname: json input file
#         """
# 
#         self.fname = fname
#         self.descr = ru.read_json (fname)
# 
# 
#     # --------------------------------------------------------------------------
#     def generate_troy_workload (self) :
# 
#         try :
#             import troy
# 
#         except Exception as e :
#             radical.synapse._logger.error ("Cannot import troy")
#             return  None
# 
# 
#         workload = troy.Workload ()
# 
#         for sequence in self.descr[ in radical_students :
#             workload.add_task (create_task_description ())
# 
#         return workload
# 
# 
# 
# with open ('%s/synapse/experiments/%s.dat' % (home, host), 'a') as f :
# 
#     _     = os.popen ('sync')
# #   _     = os.popen ('rm -f /tmp/synapse_*')
# #   _     = os.popen ('sudo sh -c "sync; echo 3 > /proc/sys/vm/drop_caches"')
# 
#     start = time.time()
# 
# 
#     load_id        = str(os.environ.get ('SYNAPSE_ID',           'X'))
#     load_instances = int(os.environ.get ('SYNAPSE_INSTANCES',      1))
#     load_compute   = int(os.environ.get ('SYNAPSE_COMPUTE_GFLOPS', 0))
#     load_memory    = int(os.environ.get ('SYNAPSE_MEMORY_GBYTES' , 0))
#     load_storage   = int(os.environ.get ('SYNAPSE_STORAGE_GBYTES', 0))
# 
#     apps = list()
# 
#     # create containers for different system workload types
#     for i in range (0, load_instances) :
# 
#         app = dict()
#         app['c'] = rsa.Compute ()
#         app['m'] = rsa.Memory  ()
#         app['s'] = rsa.Storage ()
#       # app['n'] = rsa.Network ()
# 
#         apps.append (app)
# 
# 
#     # run load (this spawns threads as quickly as possible)
#     for app in apps :
# 
#         # the atoms below are executed concurrently (in their own threads)
#         app['c'].run (info={'n'   : load_compute})  # consume  10 GFlop CPY Cycles
#         app['m'].run (info={'n'   : load_memory})   # allocate  5 GByte memory
#         app['s'].run (info={'n'   : load_storage,   # write     2 GByte to disk
#                             'tgt' : '%(tmp)s/synapse_storage.tmp.%(pid)s'})
# 
#       # app['n'].run (info={'type'   : 'server', # communicate a 1 MByte message
#       #                     'mode'   : 'read',
#       #                     'port'   : 10000,
#       #                     'n'      : 100})
#       # time.sleep (1)
#       # app['n'].run (info={'type'   : 'client',
#       #                     'mode'   : 'write',
#       #                     'host'   : 'localhost',
#       #                     'port'   : 10000,
#       #                     'n'      : 100})
# 
# 
#     # all are started -- now wait for completion and collect times
#     times = {}
#     times['c'] = 0.0
#     times['m'] = 0.0
#     times['s'] = 0.0
#   # times['n'] = 0.0
# 
#     cid = 0
#     for app in apps :
#         cid += 1
# 
#         info_c = app['c'].wait ()
#         info_m = app['m'].wait ()
#         info_s = app['s'].wait ()
#       # info_n = app['n'].wait ()
# 
#         t_c    = float(info_c['timer'])
#         t_m    = float(info_m['timer'])
#         t_s    = float(info_s['timer'])
#       # t_n    = float(info_n['timer'])
# 
#       # import pprint
#       # pprint.pprint (info_c)
# 
#         times['c'] += t_c
#         times['m'] += t_m
#         times['s'] += t_s
#       # times['n'] += t_n
# 
#         output = '%-10s %10s ------- %7.2f %7.2f %7.2f %5d %5d %5d %5d' % \
#                 (host, "%s.%002d" % (load_id, cid), t_c, t_m, t_s,
#                  load_instances, load_compute, load_memory, load_storage)
# 
#       # print output
#         f.write ("%s\n" % output)
# 
# 
#     # also print summary
#     output = '%-10s %7s    %7.2f ------- ------- ------- %5d %5d %5d %5d' % \
#              (host, load_id, time.time() - start, 
#              load_instances, load_compute, load_memory, load_storage)
# 
# #   print output
#     f.write ("%s\n" % output)
# 
#     time.sleep (10)
#     _ = os.popen ('ps -ef | grep -i "/tmp/synapse_" | grep -v grep | cut -c 8-15 | xargs -r kill -9')
# 
# #   print su.get_mem_usage ()
# #   print su.get_io_usage  ()
# 



#!/usr/bin/env python

__author__    = 'RADICAL Team'
__email__     = 'radical@rutgers.edu'
__copyright__ = 'Copyright 2013/14, RADICAL Research, Rutgers University'
__license__   = 'MIT'


""" Setup script. Used by easy_install and pip. """

import re
import os
import sys
import glob
import shutil
import subprocess as sp

name     = 'radical.synapse'
mod_root = 'src/radical/synapse/'

try:
    from setuptools import setup, Command, find_packages, Extension
except ImportError as e:
    print("%s needs setuptools to install" % name)
    sys.exit(1)

# ------------------------------------------------------------------------------
#
# versioning mechanism:
#
#   - version:          1.2.3            - is used for installation
#   - version_detail:  v1.2.3-9-g0684b06 - is used for debugging
#   - version is read from VERSION file in src_root, which then is copied to
#     module dir, and is getting installed from there.
#   - version_detail is derived from the git tag, and only available when
#     installed from git.  That is stored in mod_root/VERSION in the install
#     tree.
#   - The VERSION file is used to provide the runtime version information.
#
def get_version (mod_root):
    """
    mod_root
        a VERSION file containes the version strings is created in mod_root,
        during installation.  That file is used at runtime to get the version
        information.  
        """

    try:

        version        = None
        version_detail = None

        # get version from './VERSION'
        src_root = os.path.dirname (__file__)
        if  not src_root :
            src_root = '.'

        with open (src_root + '/VERSION', 'r') as f :
            version = f.readline ().strip()


        # attempt to get version detail information from git
        # We only do that though if we are in a repo root dir, 
        # ie. if 'git rev-parse --show-prefix' returns an empty string --
        # otherwise we get confused if the ve lives beneath another repository,
        # and the pip version used uses an install tmp dir in the ve space
        # instead of /tmp (which seems to happen with some pip/setuptools 
        # versions).
        p   = sp.Popen ('cd %s ; '\
                        'test -z `git rev-parse --show-prefix` || exit -1; '\
                        'tag=`git describe --tags --always` 2>/dev/null ; '\
                        'branch=`git branch | grep -e "^*" | cut -f 2- -d " "` 2>/dev/null ; '\
                        'echo $tag@$branch'  % src_root,
                        stdout=sp.PIPE, stderr=sp.STDOUT, shell=True)
        version_detail = str(p.communicate()[0].strip())
        version_detail = version_detail.replace('detached from ', 'detached-')

        # remove all non-alphanumeric (and then some) chars
        version_detail = re.sub('[/ ]+', '-', version_detail)
        version_detail = re.sub('[^a-zA-Z0-9_+@.-]+', '', version_detail)


        if  p.returncode   !=  0  or \
            version_detail == '@' or \
            'not-a-git-repo' in version_detail or \
            'not-found'      in version_detail or \
            'fatal'          in version_detail :
            version_detail =  version

        print('version: %s (%s)' % (version, version_detail))


        # make sure the version files exist for the runtime version inspection
        path = '%s/%s' % (src_root, mod_root)
        print('creating %s/VERSION' % path)
        with open (path + "/VERSION", "w") as f : f.write (version_detail + "\n")

        sdist_name = "%s-%s.tar.gz" % (name, version_detail)
        sdist_name = sdist_name.replace ('/', '-')
        sdist_name = sdist_name.replace ('@', '-')
        sdist_name = sdist_name.replace ('#', '-')
        sdist_name = sdist_name.replace ('_', '-')
        if '--record'  in sys.argv or 'bdist_egg' in sys.argv or 'bdist_wheel' in sys.argv:
           # pip install stage 2      easy_install stage 1
           # NOTE: pip install will untar the sdist in a tmp tree.  In that tmp
           # tree, we won't be able to derive git version tags -- so we pack the
           # formerly derived version as ./VERSION
            shutil.move ("VERSION", "VERSION.bak")           # backup version
            shutil.copy ("%s/VERSION" % path, "VERSION")     # use full version instead
            os.system   ("python setup.py sdist")            # build sdist
            shutil.copy ('dist/%s' % sdist_name,
                         '%s/%s'   % (mod_root, sdist_name)) # copy into tree
            shutil.move ("VERSION.bak", "VERSION")           # restore version

        print('creating %s/SDIST' % path)
        with open (path + "/SDIST", "w") as f : f.write (sdist_name + "\n")

        return version, version_detail, sdist_name

    except Exception as e :
        raise RuntimeError ('Could not extract/set version: %s' % e)


# ------------------------------------------------------------------------------
# check python version. we need >= 2.7, <3.x
if  sys.hexversion < 0x02070000 or sys.hexversion >= 0x03000000:
    raise RuntimeError("%s requires Python 2.x (2.7 or higher)" % name)


# ------------------------------------------------------------------------------
# get version info -- this will create VERSION and srcroot/VERSION
version, version_detail, sdist_name = get_version (mod_root)


# ------------------------------------------------------------------------------
class our_test(Command):
    user_options = []
    def initialize_options (self) : pass
    def finalize_options   (self) : pass
    def run (self) :
        testdir = "%s/tests/" % os.path.dirname(os.path.realpath(__file__))
        retval  = sp.call([sys.executable,
                          '%s/run_tests.py'               % testdir,
                          '%s/configs/basetests.cfg'      % testdir])
        raise SystemExit(retval)


# ------------------------------------------------------------------------------
#
def read(*rnames):
    try :
        return open(os.path.join(os.path.dirname(__file__), *rnames)).read()
    except Exception :
        return ''


# ------------------------------------------------------------------------------
#
# borrowed from the MoinMoin-wiki installer
#
def makeDataFiles(prefix, dir):
    """ Create distutils data_files structure from dir

    distutil will copy all file rooted under dir into prefix, excluding
    dir itself, just like 'ditto src dst' works, and unlike 'cp -r src
    dst, which copy src into dst'.

    Typical usage:
        # install the contents of 'wiki' under sys.prefix+'share/moin'
        data_files = makeDataFiles('share/moin', 'wiki')

    For this directory structure:
        root
            file1
            file2
            dir
                file
                subdir
                    file

    makeDataFiles('prefix', 'root')  will create this distutil data_files structure:
        [('prefix', ['file1', 'file2']),
         ('prefix/dir', ['file']),
         ('prefix/dir/subdir', ['file'])]

    """
    # Strip 'dir/' from of path before joining with prefix
    dir = dir.rstrip('/')
    strip = len(dir) + 1
    found = []
    os.path.walk(dir, visit, (prefix, strip, found))
    return found

def visit((prefix, strip, found), dirname, names):
    """ Visit directory, create distutil tuple

    Add distutil tuple for each directory using this format:
        (destination, [dirname/file1, dirname/file2, ...])

    distutil will copy later file1, file2, ... info destination.
    """
    files = []
    # Iterate over a copy of names, modify names
    for name in names[:]:
        path = os.path.join(dirname, name)
        # Ignore directories -  we will visit later
        if os.path.isdir(path):
            # Remove directories we don't want to visit later
            if isbad(name):
                names.remove(name)
            continue
        elif isgood(name):
            files.append(path)
    destination = os.path.join(prefix, dirname[strip:])
    found.append((destination, files))

def isbad(name):
    """ Whether name should not be installed """
    return (name.startswith('.') or
            name.startswith('#') or
            name.endswith('.pickle') or
            name == 'CVS')

def isgood(name):
    """ Whether name should be installed """
    if not isbad(name):
        if name.endswith('.py') or name.endswith('.json'):
            return True
    return False


# -------------------------------------------------------------------------------
if 'RADICAL_SYNAPSE_USE_OPENMP' in os.environ:
    eca = ['-fopenmp', '-DRADICAL_SYNAPSE_USE_OPENMP=1']
    ela = ['-lgomp']
else:
    eca = []
    ela = []

c_ext = Extension(name               = "radical/synapse/atoms/_atoms" ,
                  sources            = glob.glob('src/radical/synapse/atoms/*.c'),
                  extra_compile_args = eca, 
                  extra_link_args    = ela)

setup_args = {
    'name'               : name,
    'version'            : version,
    'description'        : 'SYNthetic APplicationS Emulator -- A RADICAL Project '
                           '(http://radical.rutgers.edu/)',
    'long_description'   : (read('README.md') + '\n\n' + read('CHANGES.md')),
    'author'             : 'RADICAL Group at Rutgers University',
    'author_email'       : 'radical@rutgers.edu',
    'maintainer'         : 'The RADICAL Group',
    'maintainer_email'   : 'radical@rutgers.edu',
    'url'                : 'https://www.github.com/radical-cybertools/radical.utils/',
    'license'            : "LGPLv3+",
    'keywords'           : "radical emulate workload",
    'classifiers'        : [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Utilities',
        'Topic :: System :: Distributed Computing',
        'Topic :: Scientific/Engineering',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],
    'ext_modules'        : [c_ext],
    'namespace_packages' : ['radical'],
    'packages'           : find_packages('src'),
    'package_dir'        : {'': 'src'},
    'scripts'            : ['bin/radical-synapse-version',
                            'bin/radical-synapse-profile',
                            'bin/radical-synapse-stats',
                            'bin/radical-synapse-stats.plot',
                            'bin/radical-synapse-execute',
                            'bin/radical-synapse-emulate',
                            'bin/radical-synapse-fixdocs',
                            'bin/radical-synapse-sample',
                            'bin/radical-synapse-mandelbrot-dummy.py',
                            'bin/radical-synapse-mandelbrot-master.py',
                            'bin/radical-synapse-mandelbrot-worker.py',
                            'bin/radical-synapse-mandelbrot-profile.py',
                            'bin/radical-synapse-mandelbrot-emulate.py',
                            'bin/radical-synapse-iotrace.sh'
                           ],
    'package_data'       : {'': ['*.sh', '*.json', '*.gz', 'VERSION', 'SDIST', sdist_name, '*.c', '*.h']},
    'cmdclass'           : {
        'test'           : our_test,
                           },
    'install_requires'   : ['psutil',
                            'radical.utils',
                            'pymongo==2.8',
                           ],
    'tests_require'      : [],
    'test_suite'         : '%s.tests' % name,
    'zip_safe'           : False,
#   'build_sphinx'       : {
#       'source-dir'     : 'docs/',
#       'build-dir'      : 'docs/build',
#       'all_files'      : 1,
#   },
#   'upload_sphinx'      : {
#       'upload-dir'     : 'docs/build/html',
#   }
    # This copies the contents of the examples/ dir under
    # sys.prefix/share/$name
    # It needs the MANIFEST.in entries to work.
    'data_files'         : makeDataFiles('share/%s/examples/' % name, 'examples'),
}

# ------------------------------------------------------------------------------

setup (**setup_args)

os.system('rm -rf src/%s.egg-info' % name)

# ------------------------------------------------------------------------------

